import pandas as pd
from skbio import read as read_sequence
from skbio import write as write_sequence
from os import mkdir, path
import warnings
import logging

from mag_annotator.summarize_vgfs import filter_to_amgs
from mag_annotator.utils import setup_logger
from mag_annotator.summarize_genomes import get_ids_from_annotations_by_row
from mag_annotator.database_handler import DatabaseHandler

# TODO: filter by taxonomic level, completeness, contamination
# TODO: filter scaffolds file, gff file
# TODO: add negate, aka pull not from given list


def get_genes_from_identifiers(annotations, genes=None, fastas=None, scaffolds=None, identifiers=None, categories=None,
                               custom_distillate=None):
    specific_genes_to_keep = list()
    # filter fastas
    if fastas is not None:
        for fasta in fastas:
            specific_genes_to_keep += list(annotations.loc[annotations['fasta'] == fasta].index)
    # filter scaffolds
    # TODO: remove this functionality or modify since scaffolds are guaranteed unique
    if scaffolds is not None:
        for scaffold in scaffolds:
            specific_genes_to_keep += list(annotations.loc[annotations['scaffold'] == scaffold].index)
    # filter genes
    if genes is not None:
        specific_genes_to_keep += genes
    # filter down annotations based on specific genes
    if len(specific_genes_to_keep) > 0:
        annotations_to_keep = annotations.loc[specific_genes_to_keep]
    else:
        annotations_to_keep = annotations

    # filter based on annotations
    if (identifiers is not None) or (categories is not None):
        annotation_genes_to_keep = list()

        # make a dictionary of genes to
        gene_to_ids = dict()
        for i, row_ids in get_ids_from_annotations_by_row(annotations_to_keep).iteritems():
           if len(row_ids) > 0:
                gene_to_ids[i] = set(row_ids)

        # get genes with ids
        if identifiers is not None:
            identifiers = set(identifiers)
            for gene, ids in gene_to_ids.items():
                if len(set(ids) & set(identifiers)) > 0:
                    annotation_genes_to_keep.append(gene)

        # get genes from distillate categories
        if categories is not None:
            database_handler = DatabaseHandler()
            genome_summary_form = pd.read_csv(database_handler.dram_sheet_locs['genome_summary_form'], sep='\t')
            if custom_distillate is not None:
                genome_summary_form = pd.concat([genome_summary_form, pd.read_csv(custom_distillate, sep='\t')])
            for level in ['module', 'sheet', 'header', 'subheader']:
                for category, frame in genome_summary_form.loc[~pd.isna(genome_summary_form[level])].groupby(level):
                    if category in categories:
                        for gene, ids in gene_to_ids.items():
                            if len(ids & set(frame['gene_id'])) > 0:
                                annotation_genes_to_keep.append(gene)
    else:
        annotation_genes_to_keep = list(annotations_to_keep.index)
    return annotation_genes_to_keep


def pull_sequences(input_tsv, input_fasta, output_fasta, fastas=None, scaffolds=None, genes=None,
                   identifiers=None, adjective_sheet:str=None, categories=None, taxonomy=None, completeness=None, contamination=None,
                   amg_flags=None, aux_scores=None, virsorter_category=None, putative_amgs=False,
                   max_auxiliary_score=3, remove_transposons=False, remove_fs=False, custom_distillate=None):

    #setup logger
    logger = logging.getLogger('strainer_log')
    setup_logger(logger)

    try:
        annotations = pd.read_csv(input_tsv, sep='\t', index_col=0)
        if adjective_sheet is not None:
            adjective_sheet = pd.read_csv(input_tsv, sep='\t', index_col=0) 
            annotations = annotations.loc[adjective_sheet.index]
        annotation_genes_to_keep = get_genes_from_identifiers(annotations, genes, fastas, scaffolds, identifiers,
                                                              categories, custom_distillate)
        annotations = annotations.loc[annotation_genes_to_keep]
        if len(annotations) == 0:
            raise ValueError("Categories or identifiers provided yielded no annotations")


        # DRAM specific filtering
        if taxonomy is not None:
            taxonomy = set(taxonomy)
            annotations = annotations.loc[[len(set(i.split(';')) & taxonomy) > 0 for i in annotations['bin_taxonomy']]]
        if completeness is not None:
            annotations = annotations.loc[annotations['bin_completeness'].astype(float) > completeness]
        if contamination is not None:
            annotations = annotations.loc[annotations['bin_contamination'].astype(float) < contamination]
        if len(annotations) == 0:
            raise ValueError("DRAM filters yielded no annotations")
        # DRAM-v specific filtering
        if putative_amgs:  # get potential amgs
            annotations = filter_to_amgs(annotations.fillna(''), max_aux=max_auxiliary_score,
                                         remove_transposons=remove_transposons, remove_fs=remove_fs)
        else:
            # filter based on virsorter categories
            if virsorter_category is not None:
                annotations = annotations.loc[[i in virsorter_category for i in annotations.virsorter]]
            # filter based on aux scores
            if aux_scores is not None:
                annotations = annotations.loc[[i in aux_scores for i in annotations.auxiliary_score]]
            # filter based on amg flags
            if amg_flags is not None:
                amg_flags = set(amg_flags)
                annotations = annotations.loc[[len(set(i) & amg_flags) > 0 if not pd.isna(i) else False
                                               for i in annotations.amg_flags]]
        if len(annotations) == 0:
            raise ValueError("DRAM-v filters yielded no annotations")

        # make output
        output_fasta_generator = (i for i in read_sequence(input_fasta, format='fasta')
                                  if i.metadata['id'] in annotations.index)
        write_sequence(output_fasta_generator, format='fasta', into=output_fasta)
    except KeyError as err:
        logger.error(err)
        logger.critical("It looks like your input files are not appropriate for the filter you are trying to apply."
                     " For example using a Viral or AMG filter on an annotations that has no viral information.")
        
        raise err
    except Exception as err:
        logger.error(err)
        logger.critical('Unknown error, exiting')
        raise err


def find_neighborhoods(annotations, genes_from_ids, distance_bp=None, distance_genes=None):
    # get neighborhoods as dataframes
    neighborhood_frames = list()
    for neighborhood_number, gene in enumerate(genes_from_ids):
        gene_row = annotations.loc[gene]
        scaffold_annotations = annotations.loc[annotations['scaffold'] == gene_row['scaffold']]
        # get neighbors based on bp
        if distance_bp is not None:
            right_dist = gene_row['end_position'] + distance_bp
            left_dist = gene_row['start_position'] - distance_bp
            neighborhood_annotations = scaffold_annotations.loc[(scaffold_annotations['end_position'] >= left_dist) &
                                                                (scaffold_annotations['start_position'] <= right_dist)]
        else:
            neighborhood_annotations = scaffold_annotations
        # get neighbors based on annotations
        if distance_genes is not None:
            right_genes = gene_row['gene_position'] + distance_genes
            left_genes = gene_row['gene_position'] - distance_genes
            neighborhood_annotations = scaffold_annotations.loc[(scaffold_annotations['gene_position'] >= left_genes) &
                                                                (scaffold_annotations['gene_position'] <= right_genes)]
        # add neighborhood number and neighborhood center as columns
        neighborhood_annotations['neighborhood_number'] = neighborhood_number
        neighborhood_annotations['neighborhood_center'] = [i == gene for i in neighborhood_annotations.index]
        neighborhood_frames.append(neighborhood_annotations)
        if len(neighborhood_annotations) == 0:
            warnings.warn("")

    # merge data frames and write to file
    return pd.concat(neighborhood_frames)


def get_gene_neighborhoods(input_file, output_dir, genes=None, identifiers=None, categories=None, genes_loc=None,
                           scaffolds_loc=None, distance_genes=None, distance_bp=None, custom_distillate=None):
    # check inputs, make output
    if distance_genes is None and distance_bp is None:
        raise ValueError("Must provide distance away in bp, genes or both.")

    # get data
    annotations = pd.read_csv(input_file, sep='\t', index_col=0)
    genes_from_ids = get_genes_from_identifiers(annotations, genes=genes, identifiers=identifiers,
                                                categories=categories, custom_distillate=custom_distillate)
    if len(genes_from_ids) == 0:
        raise ValueError("No genes were found based on your filtering parameters. No neighborhoods will be generated.")

    mkdir(output_dir)

    #setup logger
    # Get a logger
    # if log_file_path is None:
    log_file_path = path.join(output_dir, "strainer.log")
    logger = logging.getLogger('strainer_log')
    setup_logger(logger, log_file_path)
    logger.info(f"The log file is created at {log_file_path}")

    neighborhood_all_annotations = find_neighborhoods(annotations, genes_from_ids, distance_bp, distance_genes)
    neighborhood_all_annotations.to_csv(path.join(output_dir, 'neighborhood_annotations.tsv'), sep='\t')
    logger.info("Neighborhood Annotations witten to tsv")
    # filter files if given
    if genes_loc is not None:
        output_fasta_generator = (i for i in read_sequence(genes_loc, format='fasta')
                                  if i.metadata['id'] in neighborhood_all_annotations.index)
        # TODO: potentially generate one fasta file per neighborhood
        write_sequence(output_fasta_generator, format='fasta',
                       into=path.join(output_dir, 'neighborhood_genes.%s' % genes_loc.split('.')[-1]))
        logger.info("Gene Neighborhood fasta generated")
    if scaffolds_loc is not None:
        neighborhood_all_annotations['scaffold_mod'] = ['%s_%s' % (row['fasta'], row['scaffold'])
                                                        for i, row in neighborhood_all_annotations.iterrows()]
        neighborhood_scaffolds = list()
        for scaffold in read_sequence(scaffolds_loc, format='fasta'):
            if scaffold.metadata['id'] in neighborhood_all_annotations['scaffold_mod'].values:
                scaffold_frame = neighborhood_all_annotations.loc[neighborhood_all_annotations['scaffold_mod'] ==
                                                                  scaffold.metadata['id']]
                for neighborhood, neighborhood_frame in scaffold_frame.groupby('neighborhood_number'):
                    neighborhood_frame = neighborhood_frame.sort_values('start_position')
                    neighborhood_scaffolds.append(scaffold[neighborhood_frame['start_position'][0]:
                                                           neighborhood_frame['end_position'][-1]])
        write_sequence((i for i in neighborhood_scaffolds), format='fasta',
                       into=path.join(output_dir, 'neighborhood_scaffolds.fna'))
        logger.info("Scaffolds Neighborhood fasta generated")



import re
import subprocess
from os import path, stat
from urllib.request import urlopen, urlretrieve
from urllib.error import URLError
import pandas as pd
import logging
from typing import Callable

HMMSCAN_ALL_COLUMNS = ['query_id', 'query_ascession', 'query_length', 'target_id', 'target_ascession', 'target_length',
                       'full_evalue', 'full_score', 'full_bias', 'domain_number', 'domain_count', 'domain_cevalue',
                       'domain_ievalue', 'domain_score', 'domain_bias', 'target_start', 'target_end', 'alignment_start',
                       'alignment_end', 'query_start', 'query_end', 'accuracy', 'description']
HMMSCAN_COLUMN_TYPES = [str, str, int, str, str, int, float, float, float, int, int, float, float, float, float, int,
                        int, int, int, int, int, float, str]
BOUTFMT6_COLUMNS = ['qId', 'tId', 'seqIdentity', 'alnLen', 'mismatchCnt', 'gapOpenCnt', 'qStart', 'qEnd', 'tStart',
                    'tEnd', 'eVal', 'bitScore']


def download_file(url: str, output_file: str, logger: logging.Logger, alt_urls: list = None, verbose = True):
    # TODO: catching error 4 and give error message to retry or retry automatically
    links = [url] if alt_urls is None else [url] + alt_urls
    for l in links: 
        if verbose:
            print('downloading %s' % url)
        try:
            urlretrieve(l, output_file)
            return
        except BaseException as error:
            # BaseException is good http was to exact
            logger.warning(f"Something went wrong with the download of the url: {l}")
            logger.warning(error)
    raise URLError("DRAM whas not able to download a key database, check the logg for details")
    # run_process(['wget', '-O', output_file, url], verbose=verbose)


def setup_logger(logger, *log_file_paths, level=logging.INFO):
    logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(message)s')
    # create console handler
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    # create formatter and add it to the handlers
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    for log_file_path in log_file_paths:
        fh = logging.FileHandler(log_file_path)
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(fh)


def run_process(command, logger, shell:bool=False, capture_stdout:bool=True, save_output:str=None, 
                check:bool=False, stop_on_error:bool=True, verbose:bool=False) -> str:
    """
    Standardization of parameters for using subprocess.run, provides verbose mode and option to run via shell
    """
    # TODO just remove check
    try:
        results = subprocess.run(command, check=check, shell=shell,
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    except subprocess.CalledProcessError as error:
        logger.critical(f'The subcommand {command} experienced an error')
        if stop_on_error:
            raise error
    if results.returncode != 0:
        logger.critical(f'The subcommand {command} experienced an error: {results.stderr}')
        logging.debug(results.stdout)
        if stop_on_error:
           raise subprocess.SubprocessError(f"The subcommand {' '.join(command)} experienced an error, see the log for more info.")
    if save_output is not None:
        with open(save_output, 'w') as out:
            out.write(results.stdout)
    if capture_stdout:
        return results.stdout


# TODO: refactor following to methods to a shared run hmm step and individual get description steps
def parse_hmmsearch_domtblout(file):
    df_lines = list()
    for line in open(file):
        if not line.startswith('#'):
            line = line.split()
            line = line[:22] + [' '.join(line[22:])]
            df_lines.append(line)
    hmmsearch_frame = pd.DataFrame(df_lines, columns=HMMSCAN_ALL_COLUMNS)
    for i, column in enumerate(hmmsearch_frame.columns):
        hmmsearch_frame[column] = hmmsearch_frame[column].astype(HMMSCAN_COLUMN_TYPES[i])
    return hmmsearch_frame


def make_mmseqs_db(fasta_loc, output_loc, logger, create_index=True, threads=10, verbose=False):
    """Takes a fasta file and makes a mmseqs2 database for use in blast searching and hmm searching with mmseqs2"""
    run_process(['mmseqs', 'createdb', fasta_loc, output_loc], logger, verbose=verbose)
    if create_index:
        tmp_dir = path.join(path.dirname(output_loc), 'tmp')
        run_process(['mmseqs', 'createindex', output_loc, tmp_dir, '--threads', str(threads)], logger, verbose=verbose)


def multigrep(search_terms, search_against, logger, split_char='\n', output='.'):
    # TODO: multiprocess this over the list of search terms
    """Search a list of exact substrings against a database, takes name of mmseqs db index with _h to search against"""
    hits_file = path.join(output, 'hits.txt')
    with open(hits_file, 'w') as f:
        f.write('%s\n' % '\n'.join(search_terms))
    results = run_process(['grep', '-a', '-F', '-f', hits_file, search_against], logger, capture_stdout=True, verbose=False)
    processed_results = [i.strip() for i in results.strip().split(split_char)
                         if len(i) > 0]
    # remove(hits_file)
    return {i.split()[0]: i for i in processed_results if i != ''}


def merge_files(files_to_merge, outfile, has_header=False):
    """It's in the name, if has_header assumes all files have the same header"""
    with open(outfile, 'w') as outfile_handle:
        if has_header:
            outfile_handle.write(open(files_to_merge[0]).readline())
        for file in files_to_merge:
            with open(file) as f:
                if has_header:
                    _ = f.readline()
                outfile_handle.write(f.read())


def divide_chunks(l, n):
    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text  # or whatever


def remove_suffix(text, suffix):
    if text.endswith(suffix):
        return text[:-1*len(suffix)]
    return text  # or whatever


def get_ordered_uniques(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x) or pd.isna(x))]


def get_best_hits(query_db, target_db, logger, output_dir='.', query_prefix='query', target_prefix='target',
                  bit_score_threshold=60, threads=10, verbose=False):
    """Uses mmseqs2 to do a blast style search of a query db against a target db, filters to only include best hits
    Returns a file location of a blast out format 6 file with search results
    """
    # make query to target db
    tmp_dir = path.join(output_dir, 'tmp')
    query_target_db = path.join(output_dir, '%s_%s.mmsdb' % (query_prefix, target_prefix))
    run_process(['mmseqs', 'search', query_db, target_db, query_target_db, tmp_dir, '--threads', str(threads)],
                 logger, verbose=verbose)
    # filter query to target db to only best hit
    query_target_db_top = path.join(output_dir, '%s_%s.tophit.mmsdb' % (query_prefix, target_prefix))
    run_process(['mmseqs', 'filterdb', query_target_db, query_target_db_top, '--extract-lines', '1'], logger,
                verbose=verbose)
    # filter query to target db to only hits with min threshold
    query_target_db_top_filt = path.join(output_dir, '%s_%s.tophit.minbitscore%s.mmsdb'
                                         % (query_prefix, target_prefix, bit_score_threshold))
    run_process(['mmseqs', 'filterdb', '--filter-column', '2', '--comparison-operator', 'ge', '--comparison-value',
                 str(bit_score_threshold), '--threads', str(threads), query_target_db_top, query_target_db_top_filt],
                logger, verbose=verbose)
    # convert results to blast outformat 6
    forward_output_loc = path.join(output_dir, '%s_%s_hits.b6' % (query_prefix, target_prefix))
    run_process(['mmseqs', 'convertalis', query_db, target_db, query_target_db_top_filt, forward_output_loc,
                 '--threads', str(threads)], logger, verbose=verbose)
    return forward_output_loc


def get_reciprocal_best_hits(query_db, target_db, logger, output_dir='.', query_prefix='query', target_prefix='target',
                             bit_score_threshold=60, rbh_bit_score_threshold=350, threads=10, verbose=False):
    """Take results from best hits and use for a reciprocal best hits search"""
    # TODO: Make it take query_target_db as a parameter
    # create subset for second search
    query_target_db_top_filt = path.join(output_dir, '%s_%s.tophit.minbitscore%s.mmsdb'
                                         % (query_prefix, target_prefix, bit_score_threshold))  # I DON'T LIKE THIS
    query_target_db_filt_top_swapped = path.join(output_dir, '%s_%s.minbitscore%s.tophit.swapped.mmsdb'
                                                 % (query_prefix, target_prefix, bit_score_threshold))
    # swap queries and targets in results database
    run_process(['mmseqs', 'swapdb', query_target_db_top_filt, query_target_db_filt_top_swapped, '--threads',
                 str(threads)], logger, verbose=verbose)
    target_db_filt = path.join(output_dir, '%s.filt.mmsdb' % target_prefix)
    # create a subdatabase of the target database with the best hits as well as the index of the target database
    run_process(['mmseqs', 'createsubdb', query_target_db_filt_top_swapped, target_db, target_db_filt],
                logger, verbose=verbose)
    run_process(['mmseqs', 'createsubdb', query_target_db_filt_top_swapped, '%s_h' % target_db,
                 '%s_h' % target_db_filt], logger, verbose=verbose)

    return get_best_hits(target_db_filt, query_db, logger, output_dir, target_prefix, query_prefix, rbh_bit_score_threshold,
                         threads, verbose)


def run_hmmscan(genes_faa:str, db_loc:str, db_name:str, output_loc:str, formater:Callable,
                logger:logging.Logger, threads:int=2, db_handler=None, verbose:bool=False):
    output = path.join(output_loc, f'{db_name}_results.unprocessed.b6')
    run_process(['hmmsearch', '--domtblout', output, '--cpu', str(threads), db_loc, genes_faa], logger, verbose=verbose)
    # Parse hmmsearch output
    if not (path.isfile(output) and stat(output).st_size > 0):
        return pd.DataFrame()
    hits = parse_hmmsearch_domtblout(output)
    if len(hits) < 1:
        return pd.DataFrame()
    return formater(hits)


def get_sig_row(row, evalue_lim:float=1e-15):
    """Check if hmm match is significant, based on dbCAN described parameters"""
    tstart, tend, tlen, evalue = row[['target_start', 'target_end', 'target_length', 'full_evalue']].values
    perc_cov = (tend - tstart)/tlen
    if perc_cov >= .35 and evalue <= evalue_lim:
        return True
    else:
        return False


#TODO decide if we need use_hmmer_thresholds:bool=False
def generic_hmmscan_formater(hits:pd.DataFrame,  db_name:str, hmm_info_path:str=None, top_hit:bool=True):
    if hmm_info_path is None:
        hmm_info = None
        hits_sig = hits[hits.apply(get_sig_row, axis=1)]
    else:
        hmm_info = pd.read_csv(hmm_info_path, sep='\t', index_col=0)
        hits_sig = sig_scores(hits, hmm_info)
    if len(hits_sig) == 0:
        # if nothing significant then return nothing, don't get descriptions
        return pd.DataFrame()
    if top_hit:
        # Get the best hits
        hits_sig = hits_sig.sort_values('full_evalue').drop_duplicates(subset=["query_id"])
    hits_df = hits_sig[['target_id', 'query_id']]
    hits_df.set_index('query_id', inplace=True, drop=True)
    hits_df.rename_axis(None, inplace=True)
    hits_df.columns = [f"{db_name}_id"]
    if hmm_info is not None:
        hits_df = hits_df.merge(hmm_info[['definition']], how='left', left_on=f"{db_name}_id", right_index=True)
        hits_df.rename(columns={'definition': f"{db_name}_hits"}, inplace=True)
    return hits_df


def sig_scores(hits:pd.DataFrame, score_db:pd.DataFrame) -> pd.DataFrame:
    is_sig = list()
    for i, frame in hits.groupby('target_id'):
        row = score_db.loc[i]
        if row['score_type'] == 'domain':
            score = frame.domain_score
        elif row['score_type'] == 'full':
            score = frame.full_score
        elif row['score_type'] == '-':
            continue
        else:
            raise ValueError(row['score_type'])
        frame = frame.loc[score.astype(float) > float(row.threshold)]
        is_sig.append(frame)
    if len(is_sig) > 0:
        return pd.concat(is_sig)
    else:
        return pd.DataFrame()






#!/usr/bin/env python3

import argparse

from mag_annotator.annotate_vgfs import annotate_vgfs, remove_bad_chars
from mag_annotator.summarize_vgfs import summarize_vgfs
from mag_annotator.pull_sequences import pull_sequences, get_gene_neighborhoods

if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparsers = parser.add_subparsers()
    annotate_parser = subparsers.add_parser('annotate', help="Annotate viral contigs",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    distill_parser = subparsers.add_parser('distill',
                                           help="Summarize AMGs in annotated viral genome fragments",
                                           formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    strainer_parser = subparsers.add_parser('strainer', help="Strain annotations down to genes of interest",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    remove_parser = subparsers.add_parser('remove_bad_characters', help="Removes ; and = from fasta headers and "
                                                                        "VIRSorter_affi-contigs.tab files",
                                          formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    neighborhood_parser = subparsers.add_parser('neighborhoods', help="Find neighborhoods around genes of interest",
                                                formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # parser for annotating vgfs
    annotate_parser.add_argument('-i', '--input_fasta', help="fasta file, output from ", required=True)
    annotate_parser.add_argument('-v', '--virsorter_affi_contigs', help="VirSorter VIRSorter_affi-contigs.tab "
                                                                        "output file")
    annotate_parser.add_argument('-o', '--output_dir', help="output directory", required=True)
    annotate_parser.add_argument('--min_contig_size', type=int, default=2500,
                                 help='minimum contig size to be used for gene prediction')
    annotate_parser.add_argument('--split_contigs', action='store_true', default=False,
                                 help='Split contigs from input fasta into separate')
    prodigal_mode_choices = ['train', 'meta', 'single']
    annotate_parser.add_argument('--prodigal_mode', type=str, default='meta', choices=prodigal_mode_choices,
                                 help='Mode of prodigal to use for gene calling. NOTE: normal or single mode require '
                                      'genomes which are high quality with low contamination and long contigs (average '
                                      'length >3 Kbp).')
    # prodigal_trans_table_choices = ['auto'] + [str(i) for i in range(1, 26)]
    prodigal_trans_table_choices = [str(i) for i in range(1, 26)]
    annotate_parser.add_argument('--trans_table', type=str, default='11', choices=prodigal_trans_table_choices,
                                 help='Translation table for prodigal to use for gene calling')
    annotate_parser.add_argument('--bit_score_threshold', type=int, default=60,
                                 help='minimum bitScore of search to retain hits')
    annotate_parser.add_argument('--rbh_bit_score_threshold', type=int, default=350,
                                 help='minimum bitScore of reverse best hits to retain hits')
    annotate_parser.add_argument('--kofam_use_dbcan2_thresholds', action='store_true', default=False,
                                 help='Use dbcan2 suggested HMM cutoffs for KOfam annotation instead of KOfam '
                                      'recommended cutoffs. This will be ignored if annotating with KEGG Genes.')
    annotate_parser.add_argument('--custom_db_name', action='append', default=[], # empty list is not bug, can't be changed
                                 help="Names of custom databases, can be used multiple times.")
    annotate_parser.add_argument('--custom_fasta_loc', action='append', default=[], # empty list is not bug, can't be changed
                                 help="Location of fastas to annotate against, can be used multiple times but"
                                      "must match nubmer of custom_db_name's")
    annotate_parser.add_argument('--custom_hmm_name', action='append',  default=[], # empty list is not bug, can't be changed
                                 help="Names of custom hmm databases, can be used multiple times.")
    annotate_parser.add_argument('--custom_hmm_loc', action='append',  default=[], # empty list is not bug, can't be changed
                                 help="Location of hmms to annotate against, can be used multiple times but"
                                      "must match nubmer of custom_hmm_name's")
    annotate_parser.add_argument('--custom_hmm_cutoffs_loc', action='append',
                                       help="Location of file with custom HMM cutoffs and descriptions, can be used "
                                            "multiple times.")
    annotate_parser.add_argument('--use_uniref', action='store_true', default=False,
                                 help='Annotate these fastas against UniRef, drastically increases'
                                      ' run time and memory requirements')
    annotate_parser.add_argument('--low_mem_mode', action='store_true', default=False,
                                 help='Skip annotating with uniref and use kofam instead of KEGG genes even if '
                                      'provided. Drastically decreases memory usage')
    annotate_parser.add_argument('--skip_trnascan', action='store_true', default=False)
    annotate_parser.add_argument('--keep_tmp_dir', action='store_true', default=False)
    annotate_parser.add_argument('--threads', type=int, default=10, help='number of processors to use')
    annotate_parser.add_argument('--verbose', action='store_true', default=False)
    annotate_parser.add_argument('--config_loc', default=None,
                                 help='location of an alternive config file that will over write the original at run time,'
                                 ' but not be saved or modified')
    annotate_parser.set_defaults(func=annotate_vgfs)

    # parser for summarizing genomes
    distill_parser.add_argument("-i", "--input_file", help="Annotations path")
    distill_parser.add_argument("-o", "--output_dir", help="Directory to write summarized genomes", required=True)
    distill_parser.add_argument("--groupby_column", help="Column from annotations to group as VGF units",
                                default='scaffold')
    distill_parser.add_argument("--max_auxiliary_score", type=int, default=3,
                                help="Maximum auxiliary score to consider gene as potential AMG")
    distill_parser.add_argument("--remove_transposons", default=False, action='store_true',
                                help="Do not consider genes on scaffolds with transposons as potential AMGs")
    distill_parser.add_argument("--remove_fs", default=False, action='store_true',
                                help="Do not consider genes near ends of scaffolds as potential AMGs")
    distill_parser.add_argument('--log_file_path', 
                                       help="A name and loctation for the log file")
    # distill_parser.add_argument("--remove_js", default=False, action='store_true',
    #                             help="Do not consider genes on possible non-viral contigs as potential AMGs")
    distill_parser.add_argument("--custom_distillate", help="Custom distillate form to add your own modules")
    distill_parser.add_argument('--config_loc', default=None,
                                 help='location of an alternive config file that will over write the original at run time,'
                                 ' but not be saved or modified')
    distill_parser.set_defaults(func=summarize_vgfs)

    # parser for getting genes
    input_group = strainer_parser.add_argument_group('Input and output files')
    input_group.add_argument('-i', '--input_annotations', required=True, help='annotations file to pull genes from')
    input_group.add_argument('-f', '--input_fasta', required=True, help='fasta file to filter')
    input_group.add_argument('-o', '--output_fasta', default='pull_genes.fasta',
                             help='location to write filtered fasta')
    name_group = strainer_parser.add_argument_group('Specific names to keep')
    name_group.add_argument('--fastas', nargs='*', default=None, help='space separated list of fastas to keep')
    name_group.add_argument('--scaffolds', nargs='*', default=None,
                            help='space separated list of scaffolds to keep')
    name_group.add_argument('--genes', nargs='*', default=None, help='space separated list of genes to keep')
    annotation_group = strainer_parser.add_argument_group('Annotation filters')
    annotation_group.add_argument('--identifiers', nargs='*', default=None, help='database identifiers to keep')
    annotation_group.add_argument('--categories', nargs='*', default=None,
                                  help='distillate categories to keep genes from')
    dramv_group = strainer_parser.add_argument_group('DRAM-v based filters')
    dramv_group.add_argument('--amg_flags', default=None,
                             help='AMG flags to keep, if any one is present then it will be kept')
    dramv_group.add_argument('--aux_scores', nargs='*', default=None, type=int,
                             help='space separate list of auxiliary scores to keep')
    dramv_group.add_argument('--virsorter_category', nargs='*', default=None,
                             help='space separate list of virsorter categories to keep')
    amg_group = strainer_parser.add_argument_group('AMG filtering')
    amg_group.add_argument('-a', '--putative_amgs', default=False, action='store_true',
                           help='Only keep genes considered putative AMGs')
    amg_group.add_argument('--max_auxiliary_score', type=int, default=3,
                           help="Maximum auxiliary score to consider gene as potential AMG")
    amg_group.add_argument('--remove_transposons', default=False, action='store_true',
                           help="Do not consider genes on scaffolds with transposons as potential AMGs")
    amg_group.add_argument('--remove_fs', default=False, action='store_true',
                           help="Do not consider genes near ends of scaffolds as potential AMGs")
    # amg_group.add_argument("--remove_js", default=False, action='store_true',
    #                        help="Do not consider genes on possible non-viral contigs as potential AMGs")
    strainer_parser.set_defaults(func=pull_sequences)

    # parser for getting gene neighborhoods
    neighborhood_parser.add_argument("-i", "--input_file", help="Annotations path")
    neighborhood_parser.add_argument("-o", "--output_dir", help="Directory to write gene neighborhoods", required=True)
    neighborhood_parser.add_argument("--genes", nargs='*', help="Gene names from DRAM to find neighborhoods around")
    neighborhood_parser.add_argument("--identifiers", nargs='*',
                                     help="Database identifiers assigned by DRAM to find neighborhoods around")
    neighborhood_parser.add_argument("--categories", help="Distillate categories to build gene neighborhoods around.")
    neighborhood_parser.add_argument("--genes_loc", help="Location of genes.fna/genes.faa file to filter to "
                                                         "neighborhoods")
    neighborhood_parser.add_argument("--scaffolds_loc", help="Location of scaffolds.fna file to filter to "
                                                             "neighborhoods")
    neighborhood_parser.add_argument("--distance_genes", type=int, help="Number of genes away from center to include "
                                                                        "in neighborhoods")
    neighborhood_parser.add_argument("--distance_bp", type=int, help="Number of genes away from center to include "
                                                                     "in neighborhoods")
    neighborhood_parser.set_defaults(func=get_gene_neighborhoods)

    fasta_or_affi = remove_parser.add_mutually_exclusive_group(required=True)
    fasta_or_affi.add_argument('-i', '--input_fasta', help='Fasta file to remove ; and = from headers')
    fasta_or_affi.add_argument('-v', '--input_virsorter_affi_contigs', help='Fasta file to remove ; and = from headers')
    remove_parser.add_argument('-o', '--output', required=True, help='Name of output file. If cleaning a fasta file the'
                                                                     ' output file name must have no = or ;.')
    remove_parser.set_defaults(func=remove_bad_chars)

    args = parser.parse_args()
    args_dict = {i: j for i, j in vars(args).items() if i != 'func'}
    args.func(**args_dict)



import pytest

import os
import json
import pandas as pd
import logging

from mag_annotator.utils import run_process, make_mmseqs_db, \
    merge_files, multigrep, remove_prefix, remove_suffix, \
    setup_logger, parse_hmmsearch_domtblout, generic_hmmscan_formater, download_file
from mag_annotator.pull_sequences import get_genes_from_identifiers

@pytest.fixture()
def logger(tmpdir):
    logger = logging.getLogger('test_log')
    setup_logger(logger)
    return logger


def test_run_process(logger):
    run_process(['echo', 'Hello', 'World'], logger, verbose=True)
    assert True


@pytest.fixture()
def mmseqs_db_dir(tmpdir):
    output_loc = tmpdir.mkdir('make_mmseqs_db_test')
    return output_loc


@pytest.fixture()
def test_make_mmseqs_db(mmseqs_db_dir):
    faa_path = os.path.join('tests', 'data', 'NC_001422.faa')
    output_file = str(mmseqs_db_dir.join('mmseqs_db.mmsdb'))
    make_mmseqs_db(faa_path, output_file, True, 1)
    assert os.path.isfile(output_file)


@pytest.fixture()
def merge_test_dir(tmpdir):
    return tmpdir.mkdir('test_merge')


@pytest.fixture()
def files_to_merge_no_header(merge_test_dir):
    files_to_merge = list()
    for i in range(3):
        merge_file = merge_test_dir.join('merge_test.%s.txt' % str(i))
        merge_file.write('test%s\n' % str(i))
        files_to_merge.append(merge_file)
    return files_to_merge


@pytest.fixture()
def files_to_merge_w_header(merge_test_dir):
    files_to_merge = list()
    header = 'gene_name'
    for i in range(3):
        merge_file = merge_test_dir.join('merge_test_w_header.%s.txt' % str(i))
        merge_file.write('%s\ntest%s\n' % (header, str(i)))
        files_to_merge.append(merge_file)
    return files_to_merge


def test_download_file(logger, tmpdir):
    tempfile = tmpdir / 'out'
    with pytest.raises(URLError):
        download_file('https://notathingstupid.com', tempfile, logger)

def test_download_file(logger, tmpdir):
    tempfile = tmpdir / 'out'
    download_file('https://raw.githubusercontent.com/WrightonLabCSU/DRAM'
                  '/master/data/function_heatmap_form.tsv',
                  tempfile, logger)
    assert os.stat(tempfile).st_size > 0
    
def test_download_file(logger, tmpdir):
    tempfile = tmpdir / 'out'
    download_file('https://notathingstupid.nothing', tempfile, logger,
                  alt_urls=['https://raw.githubusercontent.com/'
                            'WrightonLabCSU/DRAM/master/data/'
                            'function_heatmap_form.tsv'])
    assert os.stat(tempfile).st_size > 0

def test_merge_files(files_to_merge_no_header, files_to_merge_w_header, merge_test_dir):
    test_merge = merge_test_dir.join('merged_test.txt')
    merge_files(files_to_merge_no_header, test_merge, False)
    assert len(test_merge.readlines()) == 3

    test_merge_w_header = merge_test_dir.join('merged_test_w_header.txt')
    merge_files(files_to_merge_w_header, test_merge_w_header, True)
    assert len(test_merge_w_header.readlines()) == 4


@pytest.fixture()
def multigrep_inputs(tmpdir):
    hits = ['gene1', 'gene3', 'gene5']
    data_str = "gene1 something about gene1\n" \
               "gene2 gene2 information\n" \
               "gene3 data including gene3\n" \
               "gene4 data including gene4\n" \
               "gene5 data including gene5\n"
    data_file = tmpdir.mkdir('multigrep_test').join('multigrep_test_data.txt')
    data_file.write(data_str)
    return hits, str(data_file)


def test_generic_hmmscan_formater():
    output_expt = pd.DataFrame({
        "bin_1.scaffold_1": ["K00001"],
        "bin_1.scaffold_2": ["K00002"]
    }, index=["test_id"]).T
    input_b6 = os.path.join('tests', 'data', 'unformatted_kofam.b6')
    hits = parse_hmmsearch_domtblout(input_b6)
    output_rcvd = generic_hmmscan_formater(
        hits,
        db_name='test',
        top_hit=True)
    output_rcvd.sort_index(inplace=True)
    output_expt.sort_index(inplace=True)
    assert output_rcvd.equals(output_expt), "Error in generic_hmmscan_formater"


def test_generic_hmmscan_formater_custom_cuts():
    output_expt = pd.DataFrame({
        "bin_1.scaffold_2": ["K00002", "KO2; description 2"]
    }, index=["test_id", "test_hits"]).T
    input_b6 = os.path.join('tests', 'data', 'unformatted_kofam.b6')
    hits = parse_hmmsearch_domtblout(input_b6)
    output_rcvd = generic_hmmscan_formater(
        hits,
        hmm_info_path=os.path.join('tests', 'data', 'hmm_thresholds.txt'),
        db_name='test',
        top_hit=True)
    output_rcvd.sort_index(inplace=True)
    output_expt.sort_index(inplace=True)
    assert output_rcvd.equals(output_expt), "Error in generic_hmmscan_formater, with custom cuts"


def test_multigrep(multigrep_inputs, logger):
    keys, values = multigrep_inputs
    dict_ = multigrep(keys, values, logger)
    assert len(dict_) == len(keys)
    assert dict_['gene1'] == 'gene1 something about gene1'
    assert dict_['gene3'] == 'gene3 data including gene3'
    assert dict_['gene5'] == 'gene5 data including gene5'


def test_remove_prefix():
    assert remove_prefix('prefix', 'pre') == 'fix'
    assert remove_prefix('postfix', 'pre') == 'postfix'
    text = 'loc_This'
    prefix = 'loc_'
    out_expected = 'This'
    out_real = remove_prefix(text, prefix)
    assert out_expected == out_real


def test_remove_suffix():
    assert remove_suffix('suffix', 'fix') == 'suf'
    assert remove_suffix('postfix', 'suf') == 'postfix'


@pytest.fixture()
def annotations():
    return pd.DataFrame([['bin.1', 'scaffold_1', None],
                         ['bin.1', 'scaffold_1', 'K00001'],
                         ['bin.1', 'scaffold_1', 'K00016'],
                         ['bin.1', 'scaffold_1', None],
                         ['bin.1', 'scaffold_2', None],
                         ['bin.2', 'scaffold_1', None],
                         ['bin.2', 'scaffold_1', None],
                         ['bin.2', 'scaffold_1', 'K00001,K13954']],
                        index=['gene1', 'gene2', 'gene3', 'gene4', 'gene5', 'gene6', 'gene7', 'gene8'],
                        columns=['fasta', 'scaffold', 'ko_id'])


def test_get_genes_from_identifiers(annotations):
    # filter by genes
    genes_to_keep = ['gene1', 'gene4', 'gene8']
    kept_genes = get_genes_from_identifiers(annotations, genes=genes_to_keep)
    assert set(kept_genes) == set(genes_to_keep)
    # filter by scaffolds
    scaffolds_to_keep = ['scaffold_2']
    kept_genes = get_genes_from_identifiers(annotations, scaffolds=scaffolds_to_keep)
    assert set(kept_genes) == {'gene5'}
    # filter by bin
    bins_to_keep = ['bin.2']
    kept_genes = get_genes_from_identifiers(annotations, fastas=bins_to_keep)
    assert set(kept_genes) == {'gene6', 'gene7', 'gene8'}
    # filter by bin and scaffold
    kept_genes = get_genes_from_identifiers(annotations, fastas=['bin.2'], scaffolds=['scaffold_2'])
    assert set(kept_genes) == {'gene5', 'gene6', 'gene7', 'gene8'}
    # filter by ko
    kept_genes = get_genes_from_identifiers(annotations, identifiers=['K00001'])
    assert set(kept_genes) == {'gene2', 'gene8'}
    # filter by bin and ko
    kept_genes = get_genes_from_identifiers(annotations, fastas=['bin.1'], identifiers=['K00001'])
    assert set(kept_genes) == {'gene2'}

# def test_get_gene_neighborhood
#     # test distance in bp works
#
#     # test distance in genes works
#
#     # test distances work together



import pytest

from os import path

from mag_annotator.database_setup import create_description_db


@pytest.fixture()
def db_loc(tmpdir):
    db_loc = str(tmpdir.mkdir('test_db').join('test_db.sqlite'))
    create_description_db(db_loc)
    return db_loc


def test_create_description_db(db_loc):
    assert path.isfile(db_loc)



import pytest

from os import path, environ
from sqlalchemy.orm.exc import NoResultFound
import json

from mag_annotator.database_setup import KeggDescription, create_description_db
from mag_annotator.database_handler import DatabaseHandler, print_database_locations, import_config, export_config, get_config_loc
from mag_annotator.utils import setup_logger
import logging
import pandas as pd
import numpy as np

TEST_CONF_PATH = path.join('tests', 'data', 'test_CONFIG')
TEST_CONF_PATH_1_3 = path.join('tests', 'data', 'test_CONFIG_1.3-')


@pytest.fixture()
def logger(tmpdir):
    logger = logging.getLogger('test_log')
    setup_logger(logger)
    return logger


@pytest.fixture()
def db_handler(logger):
    return DatabaseHandler(logger, TEST_CONF_PATH)


def test_db_hander_init():  # TODO: this
    pass

def test_add_descriptions_to_database(tmpdir, logger):
    db_handler = DatabaseHandler(logger, path.join('tests', 'data', 'test_CONFIG'))
    test_db = path.join(tmpdir.mkdir('test_db'), 'test_db.sqlite')
    create_description_db(test_db)
    db_handler.config['description_db'] = test_db
    db_handler.start_db_session()
    kegg_entries = [{'id': 'K00001', 'description': 'The first KO'},
                    {'id': 'K00002', 'description': 'The second KO'}]
    db_handler.add_descriptions_to_database(kegg_entries, 'kegg_description')
    assert len(db_handler.session.query(KeggDescription).all()) == 2
    # now test that clear table works
    kegg_entry = [{'id': 'K00003', 'description': 'The third KO'}]
    db_handler.add_descriptions_to_database(kegg_entry, 'kegg_description', clear_table=True)
    assert len(db_handler.session.query(KeggDescription).all()) == 1


@pytest.fixture()
def db_w_entries(tmpdir, logger):
    db_handler = DatabaseHandler(logger, path.join('tests', 'data', 'test_CONFIG'))
    test_db = path.join(tmpdir.mkdir('test_db_w_entries'), 'test_db.sqlite')
    create_description_db(test_db)
    db_handler.config['description_db'] = test_db
    db_handler.start_db_session()
    db_handler.add_descriptions_to_database([
        {'id': 'K00001', 'description': 'The first KO'},
        {'id': 'K00002', 'description': 'The second KO'}], 
                                            'kegg_description',
                                            clear_table=True)
    return db_handler


def test_get_description(db_w_entries):
    description = db_w_entries.get_description('K00001', 'kegg_description')
    assert description == 'The first KO'
    with pytest.raises(NoResultFound):
        db_w_entries.get_description('K00003', 'kegg_description')
    description_dict = db_w_entries.get_descriptions(['K00001', 'K00002'], 'kegg_description')
    assert isinstance(description_dict, dict)
    assert len(description_dict) == 2
    description_dict = db_w_entries.get_descriptions(['K00001', 'K00002', 'K00003'], 'kegg_description')
    assert isinstance(description_dict, dict)
    assert len(description_dict) == 2
    description_dict = db_w_entries.get_descriptions(['K00003'], 'kegg_description')
    assert isinstance(description_dict, dict)
    assert len(description_dict) == 0


def test_get_database_names(db_w_entries):
    names = db_w_entries.get_database_names()
    assert len(names) == 7


def test_filter_db_locs(logger):
    db_handler = DatabaseHandler(logger, path.join('tests', 'data', 'test_CONFIG'))
    db_handler.config['search_databases'] = {
        'kegg': 'a/fake/location/kegg.fasta', 
        'kofam_hmm': 'a/fake/location/kofam.hmm',
        'kofam_ko_list': 'a/fake/location/kofam_ko_list.tsv.gz',
        'vogdb': 'a/fake/location/vogdb.hmm', 
        'pfam': 'tests/data/Pfam-A_subset.full.gz',
        'dbcan': None, 'peptidase': None, 'uniref': 'a/fake/location/uniref.fasta'}
    db_handler.filter_db_locs(low_mem_mode=True)
    assert db_handler.config.get('search_databases') == {
        'kofam_hmm': 'a/fake/location/kofam.hmm',
        'kofam_ko_list': 'a/fake/location/kofam_ko_list.tsv.gz',
        'pfam': 'tests/data/Pfam-A_subset.full.gz',
        'dbcan': None, 'peptidase': None}
    db_handler.config['search_databases'] = {
        'kegg': 'a/fake/location/kegg.fasta', 
        'kofam_hmm': 'a/fake/location/kofam.hmm',
        'kofam_ko_list': 'a/fake/location/kofam_ko_list.tsv.gz',
        'vogdb': 'a/fake/location/vogdb.hmm', 
        'pfam': 'tests/data/Pfam-A_subset.full.gz',
        'dbcan': None, 'peptidase': None}
    db_handler.filter_db_locs(use_vogdb=False)
    assert db_handler.config.get('search_databases') == {
        'kegg': 'a/fake/location/kegg.fasta', 
        'kofam_hmm': 'a/fake/location/kofam.hmm',
        'kofam_ko_list': 'a/fake/location/kofam_ko_list.tsv.gz',
        'pfam': 'tests/data/Pfam-A_subset.full.gz', 
        'dbcan': None, 
        'peptidase': None}
    db_handler.config['search_databases'] = {
        'kegg': 'a/fake/location/kegg.fasta', 
        'kofam_hmm': 'a/fake/location/kofam.hmm',
        'kofam_ko_list': 'a/fake/location/kofam_ko_list.tsv.gz',
        'vogdb': 'a/fake/location/vogdb.hmm', 
        'pfam': 'tests/data/Pfam-A_subset.full.gz',
        'dbcan': None, 'peptidase': None, 
        'uniref': 'a/fake/location/uniref.fasta'}
    db_handler.filter_db_locs(use_uniref=False)
    assert db_handler.config.get('search_databases') == {
        'kegg': 'a/fake/location/kegg.fasta', 
        'kofam_hmm': 'a/fake/location/kofam.hmm',
        'kofam_ko_list': 'a/fake/location/kofam_ko_list.tsv.gz',
        'vogdb': 'a/fake/location/vogdb.hmm', 
        'pfam': 'tests/data/Pfam-A_subset.full.gz',
        'dbcan': None, 'peptidase': None}
    # old version
    db_handler.config['search_databases'] = {
        'uniref': 'a fake locations', 
        'kegg': '/a/fake/loc', 'viral': ''}
    db_handler.filter_db_locs(master_list=('uniref', 'kegg'))
    assert db_handler.config.get('search_databases') == {
        'uniref': 'a fake locations', 
        'kegg': '/a/fake/loc'}
    # test
    db_handler.config['search_databases'] = {
        'uniref': 'a fake locations', 
        'kegg': '/a/fake/loc', 'viral': ''}
    with pytest.raises(ValueError):
        db_handler.filter_db_locs(low_mem_mode=True)
    # test
    db_handler.config['search_databases'] = {
        'kofam_hmm': '1', 'kofam_ko_list': '2', 
        'kegg': '/a/fake/loc', 'viral': ''}
    db_handler.filter_db_locs(low_mem_mode=True)
    assert db_handler.config.get('search_databases') == {
        'kofam_hmm': '1', 'kofam_ko_list': '2', 'viral': ''}
    # test
    db_handler.config['search_databases'] = {
        'uniref': 'a fake locations', 'kegg': '/a/fake/loc', 'viral': ''}
    db_handler.filter_db_locs(use_uniref=True)
    assert db_handler.config.get('search_databases') == {
        'uniref': 'a fake locations', 'kegg': '/a/fake/loc', 'viral': ''}
    # test
    db_handler.config['search_databases'] = {'uniref': 'a fake locations', 'kegg': '/a/fake/loc', 'viral': ''}
    db_handler.filter_db_locs(use_uniref=False)
    assert db_handler.config.get('search_databases') == {'kegg': '/a/fake/loc', 'viral': ''}


def test_make_header_dict_from_mmseqs_db():
    test_headers = DatabaseHandler.make_header_dict_from_mmseqs_db(path.join('tests', 'data', 'NC_001422.mmsdb'))
    assert test_headers == [{'id': 'NP_040705.1', 'description': 'NP_040705.1 B [Escherichia virus phiX174]'},
                            {'id': 'NP_040704.1', 'description': 'NP_040704.1 A* [Escherichia virus phiX174]'},
                            {'id': 'NP_040703.1', 'description': 'NP_040703.1 A [Escherichia virus phiX174]'},
                            {'id': 'NP_040713.1', 'description': 'NP_040713.1 H [Escherichia virus phiX174]'},
                            {'id': 'NP_040712.1', 'description': 'NP_040712.1 G [Escherichia virus phiX174]'},
                            {'id': 'NP_040711.1', 'description': 'NP_040711.1 F [Escherichia virus phiX174]'},
                            {'id': 'NP_040710.1', 'description': 'NP_040710.1 J [Escherichia virus phiX174]'},
                            {'id': 'NP_040709.1', 'description': 'NP_040709.1 E [Escherichia virus phiX174]'},
                            {'id': 'NP_040708.1', 'description': 'NP_040708.1 D [Escherichia virus phiX174]'},
                            {'id': 'NP_040707.1', 'description': 'NP_040707.1 C [Escherichia virus phiX174]'},
                            {'id': 'NP_040706.1', 'description': 'NP_040706.1 K [Escherichia virus phiX174]'}]


def test_process_pfam_descriptions():
    description_list = DatabaseHandler.process_pfam_descriptions(path.join('tests', 'data', 'Pfam-A_subset.hmm.dat.gz'))
    assert description_list == [{'id': 'PF10417.9', 'description': 'C-terminal domain of 1-Cys peroxiredoxin'},
                                {'id': 'PF12574.8', 'description': '120 KDa Rickettsia surface antigen'},
                                {'id': 'PF09847.9', 'description': 'Membrane protein of 12 TMs'},
                                {'id': 'PF00244.20', 'description': '14-3-3 protein'},
                                {'id': 'PF16998.5', 'description': '17 kDa outer membrane surface antigen'}]


def test_process_dbcan_descriptions():
    description_list = DatabaseHandler.process_dbcan_descriptions(
        path.join('tests', 'data', 'CAZyDB.07312019.fam-activities.subset.txt'),
        path.join('tests', 'data', 'example.fam.subfam.ec.txt'),
    )
    expected = pd.DataFrame([
        {'id': 'AA0', 'description': 'AA0', 'ec': ''},
        {'id': 'AA10', 
         'description': 'AA10 (formerly CBM33) proteins are copper-dependent '
                        'lytic polysaccharide monooxygenases (LPMOs); some '
                        'proteins have been shown to act on chitin, others on '
                        'cellulose; lytic cellulose monooxygenase '
                        '(C1-hydroxylating) (EC 1.14.99.54); lytic cellulose '
                        'monooxygenase (C4-dehydrogenating)(EC 1.14.99.56); '
                        'lytic chitin monooxygenase (EC 1.14.99.53)',
         'ec': '3.2.1.78,1.14.99.53,1.14.99.54'
         },
         {'id': 'AA11', 
          'description': 'AA11 proteins are copper-dependent lytic polysaccharide '
                         'monooxygenases (LPMOs); cleavage of chitin chains with '
                         'oxidation of C-1 has been demonstrated for a AA11 LPMO '
                         'from Aspergillus oryzae;',
          'ec': '1.14.99.55'
          },
         {'id': 'AA12', 
          'description': 'AA12 The pyrroloquinoline quinone-dependent oxidoreductase '
                         'activity was demonstrated for the CC1G_09525 protein of '
                         'Coprinopsis cinerea.',
          'ec': '1.14.99.56'}
          ])
    
    assert pd.DataFrame(description_list).equals(expected)


def test_process_vogdb_descriptions():
    description_list = DatabaseHandler.process_vogdb_descriptions(
        path.join('tests', 'data', 'vog_annotations_latest.subset.tsv.gz'))
    assert description_list == [{'id': 'VOG00001', 'description': 'sp|Q5UQJ2|YR863_MIMIV Putative ankyrin repeat '
                                                                  'protein R863; Xh'},
                                {'id': 'VOG00002', 'description': 'sp|Q9J4Z6|V244_FOWPN Putative ankyrin repeat '
                                                                  'protein FPV244; Xh'},
                                {'id': 'VOG00003', 'description': 'sp|Q91FD6|388R_IIV6 Putative MSV199 '
                                                                  'domain-containing protein 388R; Xu'},
                                {'id': 'VOG00004', 'description': 'sp|P51704|RPC1_BPHC1 Repressor protein CI; Xu'},
                                {'id': 'VOG00005', 'description': 'sp|P17766|POLG_PPVNA Genome polyprotein; XhXrXs'}]


def test_export_config(tmpdir):
    test_exported_config = path.join(tmpdir.mkdir('test_import_config'), 'test_CONFIG')
    export_config(test_exported_config)
    assert isinstance(json.loads(open(test_exported_config).read()), dict)
    assert isinstance(json.loads(open(test_exported_config).read())['search_databases'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['database_descriptions'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['dram_sheets'], dict)

def test_import_config(tmpdir):
    stash_conf = path.join(tmpdir.mkdir('stash'), 'stash')
    export_config(stash_conf)
    import_config(TEST_CONF_PATH)
    test_exported_config = path.join(tmpdir.mkdir('test_import_config'), 'test_CONFIG')
    export_config(test_exported_config)
    import_config(stash_conf)
    assert isinstance(json.loads(open(test_exported_config).read()), dict)
    assert isinstance(json.loads(open(test_exported_config).read())['search_databases'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['database_descriptions'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['dram_sheets'], dict)

def test_import_config_old(tmpdir):
    stash_conf = path.join(tmpdir.mkdir('stash'), 'stash')
    export_config(stash_conf)
    import_config(TEST_CONF_PATH_1_3)
    test_exported_config = path.join(tmpdir.mkdir('test_import_config_old'), 'test_CONFIG')
    export_config(test_exported_config)
    import_config(stash_conf)
    assert isinstance(json.loads(open(test_exported_config).read()), dict)
    assert isinstance(json.loads(open(test_exported_config).read())['search_databases'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['database_descriptions'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['dram_sheets'], dict)

def test_write_config(tmpdir, db_handler):
    stash_conf = path.join(tmpdir.mkdir('stash'), 'stash')
    export_config(stash_conf)
    test_exported_config = path.join(tmpdir.mkdir('test_write_config'), 'test_CONFIG')
    db_handler.write_config(test_exported_config)
    import_config(stash_conf)
    assert isinstance(json.loads(open(test_exported_config).read()), dict)
    assert isinstance(json.loads(open(test_exported_config).read())['search_databases'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['database_descriptions'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['dram_sheets'], dict)


@pytest.fixture()
def db_handler_1_3(logger):
    return DatabaseHandler(logger, TEST_CONF_PATH_1_3)


def test_construct_from_dram_pre_1_4_0(tmpdir, db_handler_1_3):
    test_exported_config = path.join(tmpdir.mkdir('test_export'), 'test_CONFIG')
    db_handler_1_3.write_config(test_exported_config)
    assert isinstance(json.loads(open(test_exported_config).read()), dict)
    assert isinstance(json.loads(open(test_exported_config).read())['search_databases'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['database_descriptions'], dict)
    assert isinstance(json.loads(open(test_exported_config).read())['dram_sheets'], dict)


def test_print_database_locations(capsys):
    print_database_locations()
    out, err = capsys.readouterr()
    assert 'Description db: ' in out
    assert err.endswith("Logging to console\n")

def test_enviroment_variable(logger):
    mypath:str = 'my_path_that-is_unique'
    environ['DRAM_CONFIG_LOCATION'] = mypath
    assert get_config_loc() == mypath
    del(environ['DRAM_CONFIG_LOCATION'])


def test_set_database_paths(logger):
    # first test that adding nothing doesn't change CONFIG
    db_handler = DatabaseHandler(logger)
    pretest_db_dict = db_handler.config.get('search_databases')
    db_handler.set_database_paths(write_config=False)
    assert pretest_db_dict == db_handler.config.get('search_databases')
    # test that adding something that doesn't exist throws error
    test_fake_database = path.join('tests', 'data', 'fake_database.mmsdb')
    with pytest.raises(ValueError):
        db_handler.set_database_paths(kegg_loc=test_fake_database, write_config=False)
    
    # test that adding something real is really added
    kegg_loc = path.join('tests', 'data', 'fake_gff.gff')
    db_handler.set_database_paths(kegg_loc=kegg_loc, write_config=False)
    assert db_handler.config.get('search_databases')['kegg'] == path.realpath(kegg_loc)


def test_clear_config(logger):
    db_handler = DatabaseHandler(logger)
    db_handler.clear_config(write_config=False)
    assert np.all([i is None for i in db_handler.config['search_databases'].values()])
    assert np.all([i is None for i in db_handler.config['database_descriptions'].values()])
    assert db_handler.config['description_db'] is None



"""
Main control point for the annotation process
"""
import re
import io
import time
import logging
from glob import glob
from functools import partial
from datetime import datetime
from skbio.io import read as read_sequence
from skbio.io import write as write_sequence
from skbio import Sequence
from skbio.metadata import IntervalMetadata
from os import path, mkdir, stat
from shutil import rmtree, copy2
import pandas as pd

from mag_annotator.utils import (
    run_process,
    make_mmseqs_db,
    merge_files,
    multigrep,
    remove_suffix,
    setup_logger,
    run_hmmscan,
    sig_scores,
    get_sig_row,
    generic_hmmscan_formater,
    get_reciprocal_best_hits,
    get_best_hits,
    BOUTFMT6_COLUMNS,
)
from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.fasta_dup_name_test import fastas_dup_check

MAG_DBS_TO_ANNOTATE = (
    "kegg",
    "kofam_hmm",
    "kofam_ko_list",
    "uniref",
    "peptidase",
    "pfam",
    "dbcan",
    "vogdb",
)

# TODO: Bind verbose to logging
# TODO Exceptions are not fully handled by logging


def filter_fasta(fasta_loc, min_len=5000, output_loc=None):
    """Removes sequences shorter than a set minimum from fasta files, outputs an object or to a file"""
    kept_seqs = (
        seq for seq in read_sequence(fasta_loc, format="fasta") if len(seq) >= min_len
    )
    if output_loc is None:
        return list(kept_seqs)
    else:
        write_sequence(kept_seqs, format="fasta", into=output_loc)


def run_prodigal(
    fasta_loc, output_dir, logger, mode="meta", trans_table="11", verbose=False
):
    """Runs the prodigal gene caller on a given fasta file, outputs resulting files to given directory"""
    output_gff = path.join(output_dir, "genes.gff")
    output_fna = path.join(output_dir, "genes.fna")
    output_faa = path.join(output_dir, "genes.faa")

    run_process(
        [
            "prodigal",
            "-i",
            fasta_loc,
            "-p",
            mode,
            "-g",
            trans_table,
            "-f",
            "gff",
            "-o",
            output_gff,
            "-a",
            output_faa,
            "-d",
            output_fna,
        ],
        logger,
        verbose=verbose,
    )
    return output_gff, output_fna, output_faa


def process_reciprocal_best_hits(
    forward_output_loc, reverse_output_loc, target_prefix="target"
):
    """Process the forward and reverse best hits results to find reverse best hits
    Returns the query gene, target gene, if it was a reverse best hit, % identity, bit score and e-value
    """
    forward_hits = pd.read_csv(
        forward_output_loc, sep="\t", header=None, names=BOUTFMT6_COLUMNS
    )
    forward_hits = forward_hits.set_index("qId")
    reverse_hits = pd.read_csv(
        reverse_output_loc, sep="\t", header=None, names=BOUTFMT6_COLUMNS
    )
    reverse_hits = reverse_hits.set_index("qId")

    def check_hit(row: pd.Series):
        rbh = False
        if row.tId in reverse_hits.index:
            rbh = row.name == reverse_hits.loc[row.tId].tId
        return {
            "%s_hit" % target_prefix: row.tId,
            "%s_RBH" % target_prefix: rbh,
            "%s_identity" % target_prefix: row.seqIdentity,
            "%s_bitScore" % target_prefix: row.bitScore,
            "%s_eVal" % target_prefix: row.eVal,
            "index": row.name,
        }

    hits = forward_hits.apply(check_hit, axis=1, result_type="expand")
    # NOTE these lines may not be necessary
    hits.set_index("index", drop=True, inplace=True)
    hits.index.name = None
    return hits


def get_kegg_description(kegg_hits, header_dict):
    """Gets the KEGG IDs, and full KEGG hits from list of KEGG IDs for output in annotations"""
    gene_description = list()
    ko_list = list()
    for kegg_hit in kegg_hits.kegg_hit:
        header = header_dict[kegg_hit]
        gene_description.append(header)
        kos = re.findall(r"(K\d\d\d\d\d)", header)
        if len(kos) == 0:
            ko_list.append("")
        else:
            ko_list.append(",".join(kos))
    # TODO: change kegg_id to kegg_genes_id so that people get an error and not the wrong identifier
    new_df = pd.DataFrame(
        [kegg_hits["kegg_hit"].values, ko_list, gene_description],
        index=["kegg_genes_id", "ko_id", "kegg_hit"],
        columns=kegg_hits.index,
    )
    return pd.concat(
        [new_df.transpose(), kegg_hits.drop("kegg_hit", axis=1)], axis=1, sort=False
    )


def get_uniref_description(uniref_hits, header_dict):
    """Gets UniRef ID's, taxonomy and full string from list of UniRef IDs for output in annotations"""
    gene_description = list()
    uniref_list = list()
    gene_taxonomy = list()
    for uniref_hit in uniref_hits.uniref_hit:
        header = header_dict[uniref_hit]
        gene_description.append(header)
        uniref_list.append(header[header.find("RepID=") + 6 :])
        gene_taxonomy.append(re.search(r"Tax=(.*?) (\S*?)=", header).group(1))
    new_df = pd.DataFrame(
        [uniref_list, gene_description, gene_taxonomy],
        index=["uniref_id", "uniref_hit", "uniref_taxonomy"],
        columns=uniref_hits.index,
    )
    return pd.concat(
        [new_df.transpose(), uniref_hits.drop("uniref_hit", axis=1)], axis=1, sort=False
    )


def get_basic_description(hits, header_dict, db_name="viral"):
    """Get viral gene full descriptions based on headers (text before first space)"""
    hit_list = list()
    description = list()
    [hit for hit in hits["%s_hit" % db_name]]
    for hit in hits["%s_hit" % db_name]:
        header = header_dict[hit]
        hit_list.append(hit)
        description.append(header)
    new_df = pd.DataFrame(
        [hit_list, description],
        index=["%s_id" % db_name, "%s_hit" % db_name],
        columns=hits.index,
    )
    return pd.concat(
        [new_df.transpose(), hits.drop("%s_hit" % db_name, axis=1)], axis=1, sort=False
    )


def get_peptidase_description(peptidase_hits, header_dict):
    peptidase_list = list()
    peptidase_family = list()
    peptidase_descirption = list()
    for peptidase_hit in peptidase_hits.peptidase_hit:
        header = header_dict[peptidase_hit]
        peptidase_list.append(peptidase_hit)
        peptidase_family.append(re.search(r"#\w*.#", header).group()[1:-1])
        peptidase_descirption.append(header)
    new_df = pd.DataFrame(
        [peptidase_list, peptidase_family, peptidase_descirption],
        index=["peptidase_id", "peptidase_family", "peptidase_hit"],
        columns=peptidase_hits.index,
    )
    return pd.concat(
        [new_df.transpose(), peptidase_hits.drop("peptidase_hit", axis=1)],
        axis=1,
        sort=False,
    )


def run_mmseqs_profile_search(
    query_db,
    pfam_profile,
    output_loc,
    logger,
    output_prefix="mmpro_results",
    db_handler=None,
    threads=10,
    verbose=False,
):
    """Use mmseqs to run a search against pfam, currently keeping all hits and not doing any extra filtering"""
    tmp_dir = path.join(output_loc, "tmp")
    output_db = path.join(output_loc, "%s.mmsdb" % output_prefix)
    run_process(
        [
            "mmseqs",
            "search",
            query_db,
            pfam_profile,
            output_db,
            tmp_dir,
            "-k",
            "5",
            "-s",
            "7",
            "--threads",
            str(threads),
        ],
        logger,
        verbose=verbose,
    )
    output_loc = path.join(output_loc, "%s_output.b6" % output_prefix)
    run_process(
        ["mmseqs", "convertalis", query_db, pfam_profile, output_db, output_loc],
        logger,
        verbose=verbose,
    )
    pfam_results = pd.read_csv(
        output_loc, sep="\t", header=None, names=BOUTFMT6_COLUMNS
    )
    if pfam_results.shape[0] > 0:
        pfam_dict = dict()
        if db_handler is not None:
            pfam_descriptions = db_handler.get_descriptions(
                set(pfam_results.tId), "%s_description" % output_prefix
            )
        else:
            pfam_descriptions = {}
        for gene, pfam_frame in pfam_results.groupby("qId"):
            if len(pfam_descriptions) < 1:
                pfam_dict[gene] = "; ".join(pfam_frame.tId)
            else:
                pfam_dict[gene] = "; ".join(
                    [
                        "%s [%s]" % (pfam_descriptions[ascession], ascession)
                        for ascession in pfam_frame.tId
                    ]
                )
        return pd.DataFrame(pfam_dict, index=[f"{output_prefix}_hits"]).T
    else:
        return pd.DataFrame(columns=[f"{output_prefix}_hits"])


def find_best_dbcan_hit(genome: str, group: pd.DataFrame):
    group["perc_cov"] = group.apply(
        lambda x: (x["target_end"] - x["target_start"]) / x["target_length"], axis=1
    )
    group.sort_values("perc_cov", inplace=True)
    group.columns
    group.sort_values("full_evalue", inplace=True)
    return group.iloc[0]["target_id"]


def dbcan_hmmscan_formater(hits: pd.DataFrame, db_name: str, db_handler=None):
    """
    format the ouput of the dbcan database.

    Note these changes
    introduce new column for best hit from cazy database- this will be the best hit above the already established threshold (0.35 coverage, e-18 evalue) - then for the distillate pull info from the best hit only
    introduce new column for corresponding EC number information from sub families (EC numbers are subfamily ECs)
    Make sure that ids and descriptions are separate (descriptions these are family based)
    :param hits:
    :param db_name:
    :param db_handler:
    :returns:
    """
    hits_sig = hits[hits.apply(partial(get_sig_row, evalue_lim=1e-18), axis=1)]
    if len(hits_sig) == 0:
        # if nothing significant then return nothing, don't get descriptions
        return pd.DataFrame()
    hit_groups = hits_sig.groupby("query_id")
    all_hits = hit_groups.apply(
        lambda x: "; ".join(x["target_id"].apply(lambda y: y[:-4]).unique())
    )
    hits_df = pd.DataFrame(all_hits)
    hits_df.columns = [f"{db_name}_ids"]

    def description_pull(x: str):
        id_list = ([re.findall("^[A-Z]*[0-9]*", str(x))[0] for x in x.split("; ")],)
        id_list = [y for x in id_list for y in x if len(x) > 0]
        description_list = db_handler.get_descriptions(
            id_list, "dbcan_description"
        ).values()
        description_str = "; ".join(description_list)
        return description_str

    if db_handler is not None:
        hits_df[f"{db_name}_hits"] = hits_df[f"{db_name}_ids"].apply(description_pull)
        hits_df[f"{db_name}_subfam_ec"] = hits_df[f"{db_name}_ids"].apply(
            lambda x: "; ".join(
                db_handler.get_descriptions(
                    x.split("; "), "dbcan_description", description_name="ec"
                ).values()
            )
        )
    hits_df[f"{db_name}_best_hit"] = [find_best_dbcan_hit(*i) for i in hit_groups]
    hits_df.rename_axis(None, inplace=True)
    hits_df.columns
    return hits_df


def kofam_hmmscan_formater(
    hits: pd.DataFrame,
    hmm_info_path: str = None,
    use_dbcan2_thresholds: bool = False,
    top_hit: bool = True,
):
    hmm_info = pd.read_csv(hmm_info_path, sep="\t", index_col=0)
    if use_dbcan2_thresholds:
        hits_sig = hits[hits.apply(get_sig_row, axis=1)]
    else:
        hits_sig = sig_scores(hits, hmm_info)
    # if there are any significant results then parse to dataframe
    if len(hits_sig) == 0:
        return pd.DataFrame()
    kegg_dict = dict()
    for gene, frame in hits_sig.groupby("query_id"):
        # TODO: take top hit for full length genes and all hits for domains?
        # TODO: if top hit then give all e-value and bitscore info
        if top_hit:
            best_hit = frame[frame.full_evalue == frame.full_evalue.min()]
            ko_id = best_hit["target_id"].iloc[0]
            kegg_dict[gene] = [ko_id, hmm_info.loc[ko_id, "definition"]]
        else:
            kegg_dict[gene] = [
                ",".join([i for i in frame.target_id]),
                "; ".join([hmm_info.loc[i, "definition"] for i in frame.target_id]),
            ]
    return pd.DataFrame(kegg_dict, index=["ko_id", "kegg_hit"]).transpose()


def vogdb_hmmscan_formater(
    hits: pd.DataFrame, db_name: str, logger: logging.Logger, db_handler=None
):
    categories_col = f"{db_name}_categories"
    id_col = f"{db_name}_id"
    description_col = f"{db_name}_description"
    hits_sig = hits[hits.apply(get_sig_row, axis=1)]
    if len(hits_sig) == 0:
        logger.warn("No significant hits for vog_db")
        # if nothing significant then return nothing, don't get descriptions
        return pd.DataFrame(columns=[categories_col, id_col, description_col])
    # Get the best hits
    hits_best = hits_sig.sort_values("full_evalue").drop_duplicates(subset=["query_id"])
    if db_handler is None:
        hits_df = hits_best[["target_id", "query_id"]].copy()
    else:
        # get_descriptions
        desc_col = f"{db_name}_hits"
        descriptions = pd.DataFrame(
            db_handler.get_descriptions(
                hits_best["target_id"].unique(), description_col
            ),
            index=[desc_col],
        ).T
        categories = descriptions[desc_col].apply(lambda x: x.split("; ")[-1])
        descriptions[categories_col] = categories.apply(
            lambda x: ";".join(set([x[i : i + 2] for i in range(0, len(x), 2)]))
        )
        descriptions["target_id"] = descriptions.index
        hits_df = pd.merge(
            hits_best[["query_id", "target_id"]], descriptions, on=f"target_id"
        )
    hits_df.set_index("query_id", inplace=True, drop=True)
    hits_df.rename_axis(None, inplace=True)
    hits_df.rename(columns={"target_id": id_col}, inplace=True)
    return hits_df


def get_gene_data(fasta_loc):
    """Take the prodigal gene headers and get the scaffold that it came from
    Based on idba_ud 'scaffold_#' scaffold names with gene name after
    """
    df_dict = dict()
    for seq in read_sequence(fasta_loc, format="fasta"):
        split_label = seq.metadata["id"].split("_")
        scaffold = "_".join(split_label[:-1])
        gene_position = split_label[-1]
        start_position, end_position, strandedness = seq.metadata["description"].split(
            "#"
        )[1:4]
        df_dict[seq.metadata["id"]] = [
            scaffold,
            int(gene_position),
            int(start_position),
            int(end_position),
            int(strandedness),
        ]
    return pd.DataFrame.from_dict(
        df_dict,
        orient="index",
        columns=[
            "scaffold",
            "gene_position",
            "start_position",
            "end_position",
            "strandedness",
        ],
    )


def get_unannotated(fasta_loc, annotations):
    """Get the genes from the fasta which did not get any annotations"""
    return [
        seq.metadata["id"]
        for seq in read_sequence(fasta_loc, format="fasta")
        if seq.metadata["id"] not in annotations
    ]


def assign_grades(annotations):
    """Grade genes based on reverse best hits to KEGG, UniRef and Pfam"""
    grades = dict()
    for gene, row in annotations.iterrows():
        if row.get("kegg_RBH") is True:
            rank = "A"
        elif row.get("uniref_RBH") is True:
            rank = "B"
        elif not pd.isna(row.get("kegg_hit")) or not pd.isna(row.get("uniref_hit")):
            rank = "C"
        elif (
            not pd.isna(row.get("pfam_hits"))
            or not pd.isna(row.get("cazy_hits"))
            or not pd.isna(row.get("peptidase_hit"))
        ):
            rank = "D"
        else:
            rank = "E"
        grades[gene] = rank
    return pd.DataFrame(grades, index=["rank"]).T


def generate_annotated_fasta(input_fasta, annotations, verbosity="short", name=None):
    """Generates fasta entries with added annotation information to the header of a fasta
    either add best annotation (based on grade) (verbosity = short) or all annotations (verbosity = long)
    """
    for seq in read_sequence(input_fasta, format="fasta"):
        annotation = annotations.loc[seq.metadata["id"]]
        if "rank" in annotations.columns and verbosity == "short":
            annotation_str = "rank: %s" % annotation["rank"]
            if annotation["rank"] == "A":
                annotation_str += "; %s (db=%s)" % (annotation.kegg_hit, "kegg")
            elif annotation["rank"] == "B":
                annotation_str += "; %s (db=%s)" % (annotation.uniref_hit, "uniref")
            elif annotation["rank"] == "C":
                if "kegg_hit" in annotation:
                    annotation_str += "; %s (db=%s)" % (annotation.kegg_hit, "kegg")
                if "uniref_hit" in annotation:
                    annotation_str += "; %s (db=%s)" % (annotation.uniref_hit, "uniref")
            elif annotation["rank"] == "D":
                annotation_str += "; %s (db=%s)" % (annotation.pfam_hits, "pfam")
            else:
                pass
        else:
            annotation_list = []
            if "rank" in annotations.columns:
                annotation_list += ["rank: %s" % annotation["rank"]]
            if "kegg_hit" in annotations.columns:
                if not pd.isna(annotation.kegg_hit):
                    annotation_list += ["%s (db=%s)" % (annotation.kegg_hit, "kegg")]
            if "uniref_hit" in annotations.columns:
                if not pd.isna(annotation.uniref_hit):
                    annotation_list += [
                        "%s (db=%s)" % (annotation.uniref_hit, "uniref")
                    ]
            if "pfam_hits" in annotations.columns:
                if not pd.isna(annotation.pfam_hits):
                    annotation_list += ["%s (db=%s)" % (annotation.pfam_hits, "pfam")]
            if "bin_taxonomy" in annotations.columns:
                if not pd.isna(annotation.bin_taxonomy):
                    annotation_list += [annotation.bin_taxonomy]
            annotation_str = "; ".join(annotation_list)
        if name is not None:
            seq.metadata["id"] = "%s_%s" % (name, seq.metadata["id"])
        seq.metadata["description"] = annotation_str
        yield seq


def create_annotated_fasta(
    input_fasta, annotations, output_fasta, verbosity="short", name=None
):
    """For use with genes files, added annotations"""

    write_sequence(
        generate_annotated_fasta(input_fasta, annotations, verbosity, name),
        format="fasta",
        into=output_fasta,
    )


def generate_renamed_fasta(input_fasta, prefix):
    """For use with scaffolds files, merges together bins with fasta name added as a prefix to the file"""
    for seq in read_sequence(input_fasta, format="fasta"):
        seq.metadata["id"] = "%s_%s" % (prefix, seq.metadata["id"])
        yield seq


def rename_fasta(input_fasta, output_fasta, prefix):
    """See above"""
    write_sequence(
        generate_renamed_fasta(input_fasta, prefix), format="fasta", into=output_fasta
    )


# TODO: add annotations that don't end with '_id'
def annotate_gff(input_gff, output_gff, annotations, prefix=None):
    """Go through a gff and add a prefix to the scaffold and gene number for all ID's"""
    f = open(input_gff)
    o = open(output_gff, "w")
    for line in f:
        line = line.strip()
        if not line.startswith("#"):
            # replace id with new name
            old_scaffold = line.split("\t")[0]
            match = re.search(r"ID=\d*_\d*;", line)
            gene_number = match.group().split("_")[-1][:-1]
            old_gene_name = "%s_%s" % (old_scaffold, gene_number)
            if prefix is not None:  # add prefix to line name and gene name if given
                line = "%s_%s" % (prefix, line)
                gene_name = "%s_%s" % (prefix, old_gene_name)
            else:
                gene_name = old_gene_name
            line = re.sub(r"ID=\d*_\d*;", "ID=%s;" % gene_name, line)
            # get annotations to add from annotations file and add to end of line
            annotations_to_add = {
                strip_endings(i, ["_id"]): annotations.loc[old_gene_name, i]
                for i in annotations.columns
                if i.endswith("_id")
            }
            database_information = [
                'Dbxref="%s:%s"' % (key.strip(), value.strip().replace(":", "_"))
                for key, values in annotations_to_add.items()
                if not pd.isna(values)
                for value in values.split("; ")
                if value != ""
            ]
            if len(database_information) > 0:
                line += "%s;" % ";".join(database_information)
        o.write("%s\n" % line)


def make_gbk_from_gff_and_fasta(
    gff_loc="genes.gff", fasta_loc="scaffolds.fna", faa_loc="genes.faa", output_gbk=None
):
    # filter scaffolds out of fasta which are not in genes.gff
    gff_frame = pd.read_csv(gff_loc, sep="\t", comment="#", header=None)
    gff_scaffolds = set(gff_frame[0])
    capture_fasta = io.StringIO()
    write_sequence(
        (
            i
            for i in read_sequence(fasta_loc, format="fasta")
            if i.metadata["id"] in gff_scaffolds
        ),
        into=capture_fasta,
        format="fasta",
    )
    # scikit-bio can make a genbank for a concatenated gff and fasta file so make this first
    gff = open(gff_loc)
    aa_records = {i.metadata["id"]: i for i in read_sequence(faa_loc, format="fasta")}
    # the gff with fasta format is the gff then ##FASTA then the fasta
    concat_gff = "%s\n##FASTA\n%s" % (gff.read(), capture_fasta.getvalue())

    # now we can only ready one record from the gff/fasta hybrid at a time
    # read a record at a time till end of the fasta
    genbank_records = ""
    # +1 because there is no \n> for the first line in the file and +1 because we are indexing starting at 1
    for i in range(1, capture_fasta.getvalue().count("\n>") + 1 + 1):
        seq = read_sequence(
            io.StringIO(concat_gff), format="gff3", into=Sequence, seq_num=i
        )
        seq.metadata["LOCUS"] = {
            "locus_name": seq.metadata["id"],
            "size": len(seq),
            "unit": "bp",
            "mol_type": "DNA",
            "shape": "linear",
            "division": "ENV",
            "date": time.strftime("%d-%b-%Y").upper(),
        }
        seq_bounds = (
            seq.interval_metadata.lower_bound,
            seq.interval_metadata.upper_bound,
        )
        for interval in seq.interval_metadata.query([seq_bounds]):
            interval.metadata["gene"] = interval.metadata["ID"]
            if interval.metadata["ID"] in aa_records:
                interval.metadata["translation"] = aa_records[interval.metadata["ID"]]
        # need to capture genbank output so we can combine into a multi-genbank file
        capture_print = io.StringIO()
        seq.write(capture_print, format="genbank")
        genbank_records += capture_print.getvalue()

    # if no output then return as string, otherwise write write to output location
    if output_gbk is None:
        return genbank_records
    else:
        open(output_gbk, "w").write(genbank_records)


def get_dups(columns):
    keep = list()
    seen = list()
    for column in columns:
        if column in seen:
            keep.append(False)
        else:
            seen.append(column)
            keep.append(True)
    return keep


def run_trna_scan(fasta, tmp_dir, fasta_name, logger, threads=10, verbose=True):
    """Run tRNAscan-SE on scaffolds and create a table of tRNAs as a separate output"""
    raw_trnas = path.join(tmp_dir, "raw_trnas.txt")
    run_process(
        ["tRNAscan-SE", "-G", "-o", raw_trnas, "--thread", str(threads), fasta],
        logger,
        verbose=verbose,
    )
    if path.isfile(raw_trnas) and stat(raw_trnas).st_size > 0:
        trna_frame = pd.read_csv(raw_trnas, sep="\t", skiprows=[0, 2])
        trna_frame.columns = [i.strip() for i in trna_frame.columns]
        # if begin.1 or end.1 are in trnas then drop, else drop the second begin or end
        if "Begin.1" in trna_frame.columns:
            trna_frame = trna_frame.drop(["Begin.1"], axis=1)
        if "End.1" in trna_frame.columns:
            trna_frame = trna_frame.drop(["End.1"], axis=1)
        trna_frame = trna_frame.loc[:, get_dups(trna_frame.columns)]
        trna_frame.insert(0, "fasta", fasta_name)
        return trna_frame
    else:
        logger.warning("No tRNAs were detected, no trnas.tsv file will be created.")
        return None


RAW_RRNA_COLUMNS = [
    "scaffold",
    "tool_name",
    "type",
    "begin",
    "end",
    "e-value",
    "strand",
    "empty",
    "note",
]
RRNA_COLUMNS = ["fasta", "begin", "end", "strand", "type", "e-value", "note"]


def run_barrnap(fasta, fasta_name, logger, threads=10, verbose=True):
    raw_rrna_str = run_process(
        ["barrnap", "--threads", str(threads), fasta],
        logger,
        capture_stdout=True,
        check=False,
        verbose=verbose,
    )
    raw_rrna_table = pd.read_csv(
        io.StringIO(raw_rrna_str),
        skiprows=1,
        sep="\t",
        header=None,
        names=RAW_RRNA_COLUMNS,
        index_col=0,
    )
    rrna_table_rows = list()
    for gene, row in raw_rrna_table.iterrows():
        rrna_row_dict = {
            entry.split("=")[0]: entry.split("=")[1] for entry in row["note"].split(";")
        }
        rrna_table_rows.append(
            [
                fasta_name,
                row.begin,
                row.end,
                row.strand,
                rrna_row_dict["Name"].replace("_", " "),
                row["e-value"],
                rrna_row_dict.get("note", ""),
            ]
        )
    if len(raw_rrna_table) > 0:
        return pd.DataFrame(
            rrna_table_rows, index=raw_rrna_table.index, columns=RRNA_COLUMNS
        ).reset_index()
    else:
        logger.warning("No rRNAs were detected, no rrnas.tsv file will be created.")
        return None


def make_trnas_interval(scaffold, row, i):
    if row["Begin"] < row["End"]:
        begin = row["Begin"]
        end = row["End"]
        strand = "+"
    else:
        begin = row["End"]
        end = row["Begin"]
        strand = "-"
    metadata = {
        "source": "tRNAscan-SE",
        "type": "tRNA",
        "score": row["Score"],
        "strand": strand,
        "phase": 0,
        "ID": "%s_tRNA_%s" % (scaffold, i),
        "codon": row["Codon"],
        "product": "tRNA-%s" % row["Type"],
    }
    if not pd.isna(row["Note"]):
        metadata["Note"] = row["Note"]
    return begin, end, metadata


def make_rrnas_interval(scaffold, row, i):
    metadata = {
        "source": "barrnap",
        "type": "rRNA",
        "score": row["e-value"],
        "strand": row["strand"],
        "phase": 0,
        "ID": "%s_rRNA_%s" % (scaffold, i),
        "product": "%s ribosomal RNA" % row["type"].split()[0],
        "gene": row["type"],
    }
    if not pd.isna(row["note"]):
        metadata["Note"] = row["note"]
    return row["begin"], row["end"], metadata


# TODO: make it take an input and output gff location and not overwrite
# TODO: for some reason 1 is getting added to intervals when added to gff
def add_intervals_to_gff(
    annotations_loc, gff_loc, len_dict, interval_function, groupby_column, logger
):
    # get fasta length dict so we can merge, I'd love to be able to get this from somewhere else
    annotation_frame = pd.read_csv(annotations_loc, sep="\t")
    # process trnas to intervals
    annotation_dict = dict()
    for scaffold, frame in annotation_frame.groupby(groupby_column):
        if type(scaffold) is not str:
            scaffold = str(scaffold)
        else:
            scaffold = scaffold.strip()
        im = IntervalMetadata(len_dict[scaffold])
        for i, (_, row) in enumerate(frame.iterrows()):
            i += 1
            begin, end, metadata = interval_function(scaffold, row, i)
            begin -= 1
            if begin < 0:
                begin = 0
                logger.warning(
                    "Interval added to gff started less than zero, set to zero"
                )
            im.add(bounds=[(begin, end)], metadata=metadata)
        annotation_dict[scaffold] = im
    # add trna intervals to gff
    gff = list(read_sequence(gff_loc, format="gff3"))
    with open(gff_loc, "w") as f:
        for scaffold, gff_intervals in gff:
            gff_intervals = IntervalMetadata(len_dict[scaffold], gff_intervals)
            if scaffold in annotation_dict:
                gff_intervals.merge(annotation_dict[scaffold])
                gff_intervals.sort()
            f.write(
                gff_intervals.write(
                    io.StringIO(), format="gff3", seq_id=scaffold
                ).getvalue()
            )


def do_blast_style_search(
    query_db,
    target_db,
    working_dir,
    db_handler,
    formater,
    logger,
    db_name="database",
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    threads=10,
    verbose=False,
):
    """A convenience function to do a blast style reciprocal best hits search"""
    # Get kegg hits
    logger.info("Getting forward best hits from %s" % db_name)
    forward_hits = get_best_hits(
        query_db,
        target_db,
        logger,
        working_dir,
        "gene",
        db_name,
        bit_score_threshold,
        threads,
        verbose=verbose,
    )
    if stat(forward_hits).st_size == 0:
        return pd.DataFrame()
    logger.info("Getting reverse best hits from %s" % db_name)
    reverse_hits = get_reciprocal_best_hits(
        query_db,
        target_db,
        logger,
        working_dir,
        "gene",
        db_name,
        bit_score_threshold,
        rbh_bit_score_threshold,
        threads,
        verbose=verbose,
    )
    hits = process_reciprocal_best_hits(forward_hits, reverse_hits, db_name)
    logger.info("Getting descriptions of hits from %s" % (db_name))
    if "%s_description" % db_name in db_handler.get_database_names():
        header_dict = db_handler.get_descriptions(
            hits["%s_hit" % db_name], "%s_description" % db_name
        )
    else:
        header_dict = multigrep(
            hits[f"{db_name}_hit"], f"{target_db}_h", logger, "\x00", working_dir
        )
    hits = formater(hits, header_dict)
    return hits


def count_motifs(gene_faa, motif="(C..CH)"):
    motif_count_dict = dict()
    for seq in read_sequence(gene_faa, format="fasta"):
        motif_count_dict[seq.metadata["id"]] = len(list(seq.find_with_regex(motif)))
    return motif_count_dict


def strip_endings(text, suffixes: list):
    for suffix in suffixes:
        if text.endswith(suffix):
            text = text[: len(text) - len(suffix)]
    return text


def process_custom_dbs(
    custom_fasta_loc, custom_db_name, output_dir, logger, threads=1, verbose=False
):
    # if none is passed from argparse then set to tuple of len 0
    mkdir(output_dir)

    if custom_fasta_loc is None:
        custom_fasta_loc = ()
    if custom_db_name is None:
        custom_db_name = ()
    if len(custom_fasta_loc) != len(custom_db_name):
        raise ValueError(
            "Lengths of custom db fasta list and custom db name list must be the same."
        )
    custom_dbs = {
        custom_db_name[i]: custom_fasta_loc[i] for i in range(len(custom_db_name))
    }
    custom_db_locs = dict()
    for db_name, db_loc in custom_dbs.items():
        custom_db_loc = path.join(output_dir, "%s.custom.mmsdb" % db_name)
        make_mmseqs_db(db_loc, custom_db_loc, logger, threads=threads, verbose=verbose)
        custom_db_locs[db_name] = custom_db_loc
    return custom_db_locs


def process_custom_hmms(custom_hmm_loc, custom_hmm_name, logger, verbose=False):
    if custom_hmm_loc is None:
        custom_hmm_loc = ()
    if custom_hmm_name is None:
        custom_hmm_name = ()
    if len(custom_hmm_loc) != len(custom_hmm_name):
        raise ValueError(
            "Lengths of custom db hmm list and custom hmm db name list must be the same."
        )
    custom_hmm_locs = dict()
    for i in range(len(custom_hmm_name)):
        run_process(
            ["hmmpress", "-f", custom_hmm_loc[i]], logger, verbose=verbose
        )  # all are pressed just in case
        custom_hmm_locs[custom_hmm_name[i]] = custom_hmm_loc[i]
    return custom_hmm_locs


def process_custom_hmm_cutoffs(
    custom_hmm_cutoffs_loc, custom_hmm_name, logger: logging.Logger, verbose=False
):
    if custom_hmm_cutoffs_loc is None:
        return {}
    if custom_hmm_name is None:
        raise ValueError(
            "You can't use the custom_hmm_cutoffs_loc argument without the custom_hmm_name and"
            " custom_hmm_locs aguments specified."
        )
    if len(custom_hmm_cutoffs_loc) != len(custom_hmm_name):
        logger.warning(
            f"Custom hmm cutoffs and descriptions were only provided to the first {len(custom_hmm_cutoffs_loc)}."
            " The rest of the custom hmms will use standard cutoffs and have no descriptions."
        )
    return {custom_hmm_name[i]: j for i, j in enumerate(custom_hmm_cutoffs_loc)}


class Annotation:
    def __init__(
        self, name, scaffolds, genes_faa, genes_fna, gff, gbk, annotations, trnas, rrnas
    ):
        # TODO: get abspath for every input file/dir
        # TODO: check that files exist
        self.name = name
        self.scaffolds_loc = scaffolds
        self.genes_faa_loc = genes_faa
        self.genes_fna_loc = genes_fna
        self.gff_loc = gff
        self.gbk_loc = gbk
        self.annotations_loc = annotations
        self.trnas_loc = trnas
        self.rrnas_loc = rrnas

    def get_annotations(self):
        return pd.read_csv(self.annotations_loc, index_col=0, sep="\t")

    def get_trnas(self):
        return pd.read_csv(self.trnas_loc, sep="\t")

    def get_rrnas(self):
        return pd.read_csv(self.rrnas_loc, sep="\t")


def annotate_orfs(
    gene_faa,
    db_handler,
    tmp_dir,
    logger,
    custom_db_locs=(),
    custom_hmm_locs=(),
    custom_hmm_cutoffs_locs=(),
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    kofam_use_dbcan2_thresholds=False,
    threads=10,
    verbose=False,
):
    # run reciprocal best hits searches
    logger.info("Turning genes from prodigal to mmseqs2 db")
    query_db = path.join(tmp_dir, "gene.mmsdb")
    make_mmseqs_db(
        gene_faa, query_db, logger, create_index=True, threads=threads, verbose=verbose
    )

    annotation_list = list()

    if db_handler.config["search_databases"].get("kegg") is not None:
        # TODO Change the get_kegg_description name in function do_blast_style_search to formater
        # TODO think about how this can be consitent with blast and mmseqs
        annotation_list.append(
            do_blast_style_search(
                query_db,
                db_handler.config["search_databases"]["kegg"],
                tmp_dir,
                db_handler,
                get_kegg_description,
                logger,
                "kegg",
                bit_score_threshold,
                rbh_bit_score_threshold,
                threads,
                verbose,
            )
        )
    elif (
        db_handler.config["search_databases"].get("kofam_hmm") is not None
        and db_handler.config["search_databases"].get("kofam_ko_list") is not None
    ):
        logger.info("Getting hits from kofam")
        annotation_list.append(
            run_hmmscan(
                genes_faa=gene_faa,
                db_loc=db_handler.config["search_databases"]["kofam_hmm"],
                db_name="kofam_hmm",
                output_loc=tmp_dir,  # check_impliments
                threads=threads,  # check_impliments
                verbose=verbose,
                formater=partial(
                    kofam_hmmscan_formater,
                    hmm_info_path=db_handler.config["search_databases"][
                        "kofam_ko_list"
                    ],
                    top_hit=True,
                    use_dbcan2_thresholds=kofam_use_dbcan2_thresholds,
                ),
                logger=logger,
            )
        )
    else:
        logger.warning(
            "No KEGG source provided so distillation will be of limited use."
        )

    # Get uniref hits
    if db_handler.config["search_databases"].get("uniref") is not None:
        annotation_list.append(
            do_blast_style_search(
                query_db,
                db_handler.config["search_databases"]["uniref"],
                tmp_dir,
                db_handler,
                get_uniref_description,
                logger,
                "uniref",
                bit_score_threshold,
                rbh_bit_score_threshold,
                threads,
                verbose,
            )
        )

    # Get viral hits
    if db_handler.config["search_databases"].get("viral") is not None:
        get_viral_description = partial(get_basic_description, db_name="viral")
        annotation_list.append(
            do_blast_style_search(
                query_db,
                db_handler.config["search_databases"]["viral"],
                tmp_dir,
                db_handler,
                get_viral_description,
                logger,
                "viral",
                bit_score_threshold,
                rbh_bit_score_threshold,
                threads,
                verbose,
            )
        )

    # Get peptidase hits
    if db_handler.config["search_databases"].get("peptidase") is not None:
        annotation_list.append(
            do_blast_style_search(
                query_db,
                db_handler.config["search_databases"]["peptidase"],
                tmp_dir,
                db_handler,
                get_peptidase_description,
                logger,
                "peptidase",
                bit_score_threshold,
                rbh_bit_score_threshold,
                threads,
                verbose,
            )
        )

    # Get pfam hits
    if db_handler.config["search_databases"].get("pfam") is not None:
        logger.info("Getting hits from pfam")
        annotation_list.append(
            run_mmseqs_profile_search(
                query_db,
                db_handler.config["search_databases"]["pfam"],
                tmp_dir,
                logger,
                output_prefix="pfam",
                db_handler=db_handler,
                threads=threads,
                verbose=verbose,
            )
        )

    # use hmmer to detect cazy ids using dbCAN
    if db_handler.config["search_databases"].get("dbcan") is not None:
        logger.info("Getting hits from dbCAN")
        annotation_list.append(
            run_hmmscan(
                genes_faa=gene_faa,
                db_loc=db_handler.config["search_databases"]["dbcan"],
                db_name="cazy",
                output_loc=tmp_dir,
                threads=threads,
                formater=partial(
                    dbcan_hmmscan_formater, db_name="cazy", db_handler=db_handler
                ),
                logger=logger,
            )
        )

    # use hmmer to detect vogdbs
    if db_handler.config["search_databases"].get("vogdb") is not None:
        logger.info("Getting hits from VOGDB")
        annotation_list.append(
            run_hmmscan(
                genes_faa=gene_faa,
                db_loc=db_handler.config["search_databases"]["vogdb"],
                db_name="vogdb",
                threads=threads,
                output_loc=tmp_dir,
                formater=partial(
                    vogdb_hmmscan_formater,
                    db_name="vogdb",
                    logger=logger,
                    db_handler=db_handler,
                ),
                logger=logger,
            )
        )
    for db_name, db_loc in custom_db_locs.items():
        logger.info("Getting hits from %s" % db_name)
        get_custom_description = partial(get_basic_description, db_name=db_name)
        annotation_list.append(
            do_blast_style_search(
                query_db,
                db_loc,
                tmp_dir,
                db_handler,
                get_custom_description,
                logger,
                db_name,
                bit_score_threshold,
                rbh_bit_score_threshold,
                threads,
                verbose,
            )
        )

    # get hits to hmm style custom databases
    for hmm_name, hmm_loc in custom_hmm_locs.items():
        annotation_list.append(
            run_hmmscan(
                genes_faa=gene_faa,
                db_loc=hmm_loc,
                db_name=hmm_name,
                threads=threads,
                output_loc=tmp_dir,
                formater=partial(
                    generic_hmmscan_formater,
                    db_name=hmm_name,
                    hmm_info_path=custom_hmm_cutoffs_locs.get(hmm_name),
                    top_hit=True,
                ),
                logger=logger,
            )
        )

    annotation_list.append(
        pd.DataFrame(
            count_motifs(gene_faa, "(C..CH)"), index=["heme_regulatory_motif_count"]
        ).T
    )

    # merge dataframes
    logger.info("Merging ORF annotations")
    annotations = pd.concat(annotation_list, axis=1, sort=False)

    # get scaffold data and assign grades
    grades = assign_grades(annotations)
    annotations = pd.concat([grades, annotations], axis=1, sort=False)

    return annotations


def annotate_fasta(
    fasta_loc,
    fasta_name,
    output_dir,
    db_handler,
    logger,
    min_contig_size=5000,
    prodigal_mode="meta",
    trans_table="11",
    custom_db_locs=(),
    custom_hmm_locs=(),
    custom_hmm_cutoffs_locs=(),
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    kofam_use_dbcan2_thresholds=False,
    skip_trnascan=False,
    threads=1,
    rename_bins=True,
    keep_tmp_dir=False,
    verbose=False,
):
    """Annotated a single multifasta file, all file based outputs will be in output_dir"""
    # make temporary directory
    tmp_dir = path.join(output_dir, "tmp")
    mkdir(tmp_dir)

    # filter input fasta
    filtered_fasta = path.join(tmp_dir, "filtered_fasta.fa")
    filter_fasta(fasta_loc, min_contig_size, filtered_fasta)

    if stat(filtered_fasta).st_size == 0:
        logger.warning("No sequences were longer than min_contig_size")
        return None
    # predict ORFs with prodigal
    # TODO: handle when prodigal returns no genes
    gene_gff, gene_fna, gene_faa = run_prodigal(
        filtered_fasta,
        tmp_dir,
        logger,
        mode=prodigal_mode,
        trans_table=trans_table,
        verbose=verbose,
    )

    # annotate ORFs
    annotations = annotate_orfs(
        gene_faa,
        db_handler,
        tmp_dir,
        logger,
        custom_db_locs,
        custom_hmm_locs,
        custom_hmm_cutoffs_locs,
        bit_score_threshold,
        rbh_bit_score_threshold,
        kofam_use_dbcan2_thresholds,
        threads,
        verbose,
    )
    annotations = pd.concat([get_gene_data(gene_faa), annotations], axis=1, sort=False)

    renamed_scaffolds = path.join(output_dir, "scaffolds.annotated.fa")
    if rename_bins:
        # rename scaffolds to match prodigal names
        prefix = fasta_name
        rename_fasta(filtered_fasta, renamed_scaffolds, prefix=prefix)
    else:
        prefix = None
        copy2(filtered_fasta, renamed_scaffolds)

    # generate fna and faa output files with annotations
    annotated_fna = path.join(output_dir, "genes.annotated.fna")
    create_annotated_fasta(gene_fna, annotations, annotated_fna, name=prefix)
    annotated_faa = path.join(output_dir, "genes.annotated.faa")
    create_annotated_fasta(gene_faa, annotations, annotated_faa, name=prefix)

    # rename gff entries to match prodigal names
    renamed_gffs = path.join(output_dir, "genes.annotated.gff")
    annotate_gff(gene_gff, renamed_gffs, annotations, prefix=prefix)

    # add fasta name to frame and index, append to list
    annotations.insert(0, "fasta", fasta_name)
    if rename_bins:
        annotations.index = annotations.fasta + "_" + annotations.index
    annotations_loc = path.join(output_dir, "annotations.tsv")
    annotations.to_csv(annotations_loc, sep="\t")

    # get tRNAs and rRNAs
    len_dict = {
        i.metadata["id"]: len(i)
        for i in read_sequence(renamed_scaffolds, format="fasta")
    }
    if not skip_trnascan:
        trna_table = run_trna_scan(
            renamed_scaffolds,
            tmp_dir,
            fasta_name,
            logger,
            threads=threads,
            verbose=verbose,
        )
        if trna_table is not None:
            trna_loc = path.join(output_dir, "trnas.tsv")
            trna_table.to_csv(trna_loc, sep="\t", index=False)
            add_intervals_to_gff(
                trna_loc, renamed_gffs, len_dict, make_trnas_interval, "Name", logger
            )
        else:
            trna_loc = None
    else:
        trna_loc = None

    rrna_table = run_barrnap(
        renamed_scaffolds, fasta_name, logger, threads=threads, verbose=verbose
    )
    if rrna_table is not None:
        rrna_loc = path.join(output_dir, "rrnas.tsv")
        rrna_table.to_csv(rrna_loc, sep="\t", index=False)
        add_intervals_to_gff(
            rrna_loc, renamed_gffs, len_dict, make_rrnas_interval, "scaffold", logger
        )
    else:
        rrna_loc = None

    # make genbank file
    current_gbk = path.join(output_dir, "%s.gbk" % fasta_name)
    make_gbk_from_gff_and_fasta(
        renamed_gffs, renamed_scaffolds, annotated_faa, current_gbk
    )

    if not keep_tmp_dir:
        rmtree(tmp_dir)

    return Annotation(
        name=fasta_name,
        scaffolds=renamed_scaffolds,
        genes_faa=annotated_faa,
        genes_fna=annotated_fna,
        gff=renamed_gffs,
        gbk=current_gbk,
        annotations=annotations_loc,
        trnas=trna_loc,
        rrnas=rrna_loc,
    )


def get_fasta_name(fasta_loc):
    return path.splitext(path.basename(remove_suffix(fasta_loc, ".gz")))[0]


def annotate_fastas(
    fasta_locs,
    output_dir,
    db_handler,
    logger,
    min_contig_size=5000,
    prodigal_mode="meta",
    trans_table="11",
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    custom_db_name=(),
    custom_fasta_loc=(),
    custom_hmm_name=(),
    custom_hmm_loc=(),
    custom_hmm_cutoffs_loc=(),
    kofam_use_dbcan2_thresholds=False,
    skip_trnascan=False,
    rename_bins=True,
    keep_tmp_dir=True,
    threads=10,
    verbose=True,
):
    # check for no conflicting options/configurations
    tmp_dir = path.join(output_dir, "working_dir")
    mkdir(tmp_dir)

    # setup custom databases to be searched
    custom_db_locs = process_custom_dbs(
        custom_fasta_loc,
        custom_db_name,
        path.join(tmp_dir, "custom_dbs"),
        logger,
        threads,
        verbose,
    )
    custom_hmm_locs = process_custom_hmms(custom_hmm_loc, custom_hmm_name, logger)
    custom_hmm_cutoffs_locs = process_custom_hmm_cutoffs(
        custom_hmm_cutoffs_loc, custom_hmm_name, logger
    )
    logger.info("Retrieved database locations and descriptions")

    # iterate over list of fastas and annotate each individually
    annotations_list = list()
    for fasta_loc in fasta_locs:
        # get name of file e.g. /home/shaffemi/my_genome.fa -> my_genome
        fasta_name = get_fasta_name(fasta_loc)
        logger.info("Annotating %s" % fasta_name)
        fasta_dir = path.join(tmp_dir, fasta_name)
        mkdir(fasta_dir)
        annotations_list.append(
            annotate_fasta(
                fasta_loc,
                fasta_name,
                fasta_dir,
                db_handler,
                logger,
                min_contig_size,
                prodigal_mode,
                trans_table,
                custom_db_locs,
                custom_hmm_locs,
                custom_hmm_cutoffs_locs,
                bit_score_threshold,
                rbh_bit_score_threshold,
                kofam_use_dbcan2_thresholds,
                skip_trnascan,
                threads,
                rename_bins,
                keep_tmp_dir,
                verbose,
            )
        )
    logger.info("Annotations complete, processing annotations")

    all_annotations = merge_annotations(annotations_list, output_dir)

    # clean up
    if not keep_tmp_dir:
        rmtree(tmp_dir)
    return all_annotations


def make_fasta_namses_df(fasta_loc):
    fasta_name = get_fasta_name(fasta_loc)
    names = [
        {"fasta": fasta_name, "seq": seq.metadata["id"]}
        for seq in read_sequence(fasta_loc, format="fasta")
    ]
    return pd.DataFrame(names)


# TODO: Add force flag to remove output dir if it already exists
# TODO: Add continute flag to continue if output directory already exists
# TODO: make fasta loc either a string or list to remove annotate_bins_cmd and annotate_called_genes_cmd?
def annotate_bins(
    input_fasta: list,
    output_dir=".",
    min_contig_size=2500,
    prodigal_mode="meta",
    trans_table="11",
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    custom_db_name=(),
    custom_fasta_loc=(),
    custom_hmm_name=(),
    custom_hmm_loc=(),
    custom_hmm_cutoffs_loc=(),
    use_uniref=False,
    use_vogdb=False,
    kofam_use_dbcan2_thresholds=False,
    skip_trnascan=False,
    gtdb_taxonomy=(),
    checkm_quality=(),
    rename_bins=True,
    keep_tmp_dir=True,
    low_mem_mode=False,
    threads=10,
    verbose=True,
    log_file_path: str = None,
    join_fastas: bool = False,
    config_loc: str = None,
):
    rename_bins = True
    fasta_locs = [j for i in input_fasta for j in glob(i)]
    mkdir(output_dir)
    if log_file_path is None:
        log_file_path = path.join(output_dir, "annotate.log")
    logger = logging.getLogger("annotation_log")
    setup_logger(logger, log_file_path)
    logger.info(f"The log file is created at {log_file_path}.")

    if len(fasta_locs) == 0:
        raise ValueError("Given fasta locations return no paths: %s" % input_fasta)
    fasta_names = [get_fasta_name(i) for i in fasta_locs]
    if len(fasta_names) != len(set(fasta_names)):
        raise ValueError(
            "Genome file names must be unique. At least one name appears twice in this search."
        )
    logger.info("%s FASTAs found" % len(fasta_locs))
    # set up
    db_handler = DatabaseHandler(logger, config_loc)
    db_handler.filter_db_locs(
        low_mem_mode,
        use_uniref,
        use_vogdb,
        master_list=MAG_DBS_TO_ANNOTATE,
    )
    db_conf = db_handler.get_settings_str()
    logger.info(
        f"Starting the Annotation of Bins with database configuration: \n {db_conf}"
    )

    # check inputs
    prodigal_modes = ["train", "meta", "single"]
    if prodigal_mode not in prodigal_modes:
        raise ValueError("Prodigal mode must be one of %s." % ", ".join(prodigal_modes))
    elif prodigal_mode in ["normal", "single"]:
        logger.warning(
            "When running prodigal in single mode your bins must have long contigs (average length >3 Kbp), "
            "be long enough (total length > 500 Kbp) and have very low contamination in order for prodigal "
            "training to work well."
        )

    # prodigal_trans_tables = ['auto'] + [str(i) for i in range(1, 26)]
    prodigal_trans_tables = [str(i) for i in range(1, 26)]
    if trans_table not in prodigal_trans_tables:
        # raise ValueError('Prodigal translation table must be 1-25 or auto')
        raise ValueError("Prodigal translation table must be 1-25")

    all_annotations = annotate_fastas(
        fasta_locs,
        output_dir,
        db_handler,
        logger,
        min_contig_size,
        prodigal_mode,
        trans_table,
        bit_score_threshold,
        rbh_bit_score_threshold,
        custom_db_name,
        custom_fasta_loc,
        custom_hmm_name,
        custom_hmm_loc,
        custom_hmm_cutoffs_loc,
        kofam_use_dbcan2_thresholds,
        skip_trnascan,
        rename_bins,
        keep_tmp_dir,
        threads,
        verbose,
    )
    # if given add taxonomy information
    if len(gtdb_taxonomy) > 0:
        gtdb_taxonomy = pd.concat(
            [pd.read_csv(i, sep="\t", index_col=0) for i in gtdb_taxonomy]
        )
        taxonomy = list()
        taxonomy_missing_bins = list()
        for i in all_annotations.fasta:
            # add taxonomy
            if i in gtdb_taxonomy.index:
                taxonomy.append(gtdb_taxonomy.loc[i, "classification"])
            else:
                taxonomy.append(i)
                taxonomy_missing_bins.append(i)
        for i in set(taxonomy_missing_bins):
            logger.warning(
                "Bin %s was not found in taxonomy file, replaced with bin name." % i
            )
        all_annotations["bin_taxonomy"] = taxonomy
    # if given add quality information
    if len(checkm_quality) > 0:
        checkm_quality = pd.concat(
            [pd.read_csv(i, sep="\t", index_col=0) for i in checkm_quality]
        )
        checkm_quality.index = [
            strip_endings(i, [".fa", ".fasta", ".fna"]) for i in checkm_quality.index
        ]

        completeness = list()
        contamination = list()
        quality_missing_bins = list()
        for i in all_annotations.fasta:
            # add completeness and contamination
            if i in checkm_quality.index:
                completeness.append(checkm_quality.loc[i, "Completeness"])
                contamination.append(checkm_quality.loc[i, "Contamination"])
            else:
                completeness.append(0)
                contamination.append(100)
                quality_missing_bins.append(i)
            for j in set(quality_missing_bins):
                logger.warning(
                    "Bin %s was not found in quality file, "
                    "replaced with completeness 0 and contamination 100." % j
                )
        all_annotations["bin_completeness"] = completeness
        all_annotations["bin_contamination"] = contamination
    all_annotations.to_csv(path.join(output_dir, "annotations.tsv"), sep="\t")

    logger.info("Completed annotations")


def annotate_called_genes_cmd(
    input_faa,
    output_dir=".",
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    custom_db_name=(),
    custom_fasta_loc=(),
    custom_hmm_loc=(),
    custom_hmm_name=(),
    custom_hmm_cutoffs_loc=(),
    use_uniref=False,
    log_file_path: str = None,
    use_vogdb=False,
    kofam_use_dbcan2_thresholds=False,
    rename_genes=True,
    keep_tmp_dir=True,
    low_mem_mode=False,
    threads=10,
    verbose=True,
    config_loc: str = None,
):
    fasta_locs = glob(input_faa)
    annotate_called_genes(
        fasta_locs,
        output_dir,
        bit_score_threshold,
        rbh_bit_score_threshold,
        custom_db_name,
        custom_fasta_loc,
        custom_hmm_loc,
        custom_hmm_name,
        custom_hmm_cutoffs_loc,
        use_uniref,
        use_vogdb,
        kofam_use_dbcan2_thresholds,
        rename_genes,
        keep_tmp_dir,
        low_mem_mode,
        threads,
        verbose,
        None,
        config_loc,
    )


def perform_fasta_checks(fasta_locs, logger):
    """Perform all checks related to integrity of fastas"""
    if len(fasta_locs) == 0:
        raise ValueError("Given fasta locations returns no paths: %s" % input_faa)
    logger.info("%s fastas found" % len(fasta_locs))
    logger.info("Checking for duplicate names")
    names = [get_fasta_name(i) for i in fasta_locs]
    try:
        if len(names) != len(set(names)):
            raise ValueError(
                "Genome file names must be unique. At least one name appears twice in this search."
            )
        fastas_dup_check(fasta_locs, ">")
    except ValueError as error:
        logger.critical(error)
        raise error
    logger.info("No duplicate names found")


def annotate_called_genes(
    fasta_locs,
    output_dir=".",
    bit_score_threshold=60,
    rbh_bit_score_threshold=350,
    custom_db_name=(),
    custom_fasta_loc=(),
    custom_hmm_loc=(),
    custom_hmm_name=(),
    custom_hmm_cutoffs_loc=(),
    use_uniref=False,
    use_vogdb=False,
    kofam_use_dbcan2_thresholds=False,
    rename_genes=True,
    keep_tmp_dir=True,
    low_mem_mode=False,
    threads=10,
    verbose=True,
    log_file_path: str = None,
    config_loc: str = None,
):
    mkdir(output_dir)

    # Get a logger
    if log_file_path is None:
        log_file_path = path.join(output_dir, "annotate.log")
    logger = logging.getLogger("annotation_log")
    setup_logger(logger, log_file_path)
    logger.info(f"The log file is created at {log_file_path}")
    # get database locations
    db_handler = DatabaseHandler(logger, config_loc)
    db_handler.filter_db_locs(
        low_mem_mode, use_uniref, use_vogdb, master_list=MAG_DBS_TO_ANNOTATE
    )

    # Check fastas
    perform_fasta_checks(fasta_locs, logger)

    logger.info(
        f"Starting the Annotation of Genes with database configuration: \n {db_handler.get_settings_str()}"
    )

    logger.info("Checking for duplicate names")
    fasta_names = [get_fasta_name(i) for i in fasta_locs]
    if len(fasta_names) != len(set(fasta_names)):
        raise ValueError(
            "Genome file names must be unique. At least one name appears twice in this search."
        )
    logger.info("No duplicate names found")

    tmp_dir = path.join(output_dir, "working_dir")
    mkdir(tmp_dir)

    # setup custom databases to be searched
    custom_db_locs = process_custom_dbs(
        custom_fasta_loc,
        custom_db_name,
        path.join(tmp_dir, "custom_dbs"),
        logger,
        threads,
        verbose,
    )
    custom_hmm_locs = process_custom_hmms(custom_hmm_loc, custom_hmm_name, logger)
    custom_hmm_cutoffs_locs = process_custom_hmm_cutoffs(
        custom_hmm_cutoffs_loc, custom_hmm_name, logger
    )
    logger.info("Retrieved database locations and descriptions")

    # annotate
    annotation_locs = list()
    faa_locs = list()
    for fasta_loc in fasta_locs:
        # set up
        fasta_name = get_fasta_name(fasta_loc)
        fasta_dir = path.join(tmp_dir, fasta_name)
        mkdir(fasta_dir)

        # annotate
        annotations = annotate_orfs(
            fasta_loc,
            db_handler,
            fasta_dir,
            logger,
            custom_db_locs,
            custom_hmm_locs,
            custom_hmm_cutoffs_locs,
            bit_score_threshold,
            rbh_bit_score_threshold,
            kofam_use_dbcan2_thresholds,
            threads,
            verbose,
        )

        annotated_faa = path.join(fasta_dir, "genes.faa")

        create_annotated_fasta(fasta_loc, annotations, annotated_faa, name=fasta_name)
        faa_locs.append(annotated_faa)

        # add fasta name to frame and index, write file
        annotations.insert(0, "fasta", fasta_name)
        if rename_genes:
            annotations.index = annotations.fasta + "_" + annotations.index
        annotation_loc = path.join(fasta_dir, "annotations.tsv")
        annotations.to_csv(annotation_loc, sep="\t")
        annotation_locs.append(annotation_loc)

    # merge
    all_annotations = pd.concat(
        [pd.read_csv(i, sep="\t", index_col=0) for i in annotation_locs], sort=False
    )
    all_annotations = all_annotations.sort_values("fasta")
    all_annotations.to_csv(path.join(output_dir, "annotations.tsv"), sep="\t")
    merge_files(faa_locs, path.join(output_dir, "genes.faa"))

    # clean up
    if not keep_tmp_dir:
        rmtree(tmp_dir)

    logger.info("Completed annotations")


def merge_annotations(annotations_list, output_dir, write_annotations=False):

    # merge annotation dicts
    all_annotations = pd.concat(
        [i.get_annotations() for i in annotations_list if i is not None], sort=False
    )
    all_annotations = all_annotations.sort_values(
        ["fasta", "scaffold", "gene_position"]
    )

    # merge gene files
    merge_files(
        [i.genes_fna_loc for i in annotations_list if i is not None],
        path.join(output_dir, "genes.fna"),
    )
    merge_files(
        [i.genes_faa_loc for i in annotations_list if i is not None],
        path.join(output_dir, "genes.faa"),
    )
    merge_files(
        [i.scaffolds_loc for i in annotations_list if i is not None],
        path.join(output_dir, "scaffolds.fna"),
    )
    merge_files(
        [i.gff_loc for i in annotations_list if i is not None],
        path.join(output_dir, "genes.gff"),
        True,
    )
    trnas_locs = [
        i.trnas_loc
        for i in annotations_list
        if i is not None
        if i.trnas_loc is not None
    ]
    if len(trnas_locs) > 0:
        merge_files(trnas_locs, path.join(output_dir, "trnas.tsv"), True)
    rrnas_locs = [
        i.rrnas_loc
        for i in annotations_list
        if i is not None
        if i.rrnas_loc is not None
    ]
    if len(rrnas_locs) > 0:
        merge_files(rrnas_locs, path.join(output_dir, "rrnas.tsv"), True)

    # make output gbk dir
    gbk_dir = path.join(output_dir, "genbank")
    mkdir(gbk_dir)
    for anno in annotations_list:
        if anno is not None:
            # TODO: make annotate_fasta generate a genbank dir and then copy it's contents, get rid of Annotation.name
            if path.isfile(anno.gbk_loc):
                copy2(anno.gbk_loc, path.join(gbk_dir, "%s.gbk" % anno.name))
            else:
                for gbk_loc in glob(path.join(anno.gbk_loc, "*.gbk")):
                    copy2(gbk_loc, gbk_dir)
    if write_annotations:
        all_annotations.to_csv(path.join(output_dir, "annotations.tsv"), sep="\t")
    else:
        return all_annotations


def merge_annotations_cmd(input_dirs, output_dir):

    mkdir(output_dir)
    # Get a logger
    annotations_list = list()
    log_file_path = path.join(output_dir, "annotate.log")
    logger = logging.getLogger("annotate.log")
    setup_logger(logger, log_file_path)
    logger.info(f"The log file is created at {log_file_path}")

    # make Annotation objects per directory
    for i, annotation_dir in enumerate(glob(input_dirs)):
        if not path.exists(path.join(annotation_dir, "annotations.tsv")):
            logger.warning(
                "Skipping the fallowing directory because"
                f" no 'annotations.tsv' file was found: {annotation_dir}"
            )
            continue
        name = "annotation_%s" % i
        scaffolds = path.join(annotation_dir, "scaffolds.fna")
        genes_faa = path.join(annotation_dir, "genes.faa")
        genes_fna = path.join(annotation_dir, "genes.fna")
        gff = path.join(annotation_dir, "genes.gff")
        gbk = path.join(annotation_dir, "genbank")
        annotations = path.join(annotation_dir, "annotations.tsv")
        trnas = path.join(annotation_dir, "trnas.tsv")
        if not path.isfile(trnas):
            logger.warning("No trnas.tsv file found in directory %s" % annotation_dir)
            trnas = None
        rrnas = path.join(annotation_dir, "rrnas.tsv")
        if not path.isfile(rrnas):
            logger.warning("No rrnas.tsv file found in directory %s" % annotation_dir)
            rrnas = None
        annotations_list.append(
            Annotation(
                name=name,
                scaffolds=scaffolds,
                genes_faa=genes_faa,
                genes_fna=genes_fna,
                gff=gff,
                gbk=gbk,
                annotations=annotations,
                trnas=trnas,
                rrnas=rrnas,
            )
        )
    # run merge_annotations
    merge_annotations(annotations_list, output_dir, write_annotations=True)



#!/usr/bin/env python3

import argparse

from mag_annotator.annotate_bins import annotate_bins, annotate_called_genes_cmd, merge_annotations_cmd
from mag_annotator.summarize_genomes import summarize_genomes
from mag_annotator.pull_sequences import pull_sequences, get_gene_neighborhoods

# TODO: refactor parsers to limit duplication

if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparsers = parser.add_subparsers()
    annotate_parser = subparsers.add_parser('annotate', help="Annotate genomes/contigs/bins/MAGs",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    annotate_genes_parser = subparsers.add_parser('annotate_genes', help="Annotate already called genes, limited "
                                                                         "functionality compared to annotate",
                                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    distill_parser = subparsers.add_parser('distill', help="Summarize metabolic content of annotated genomes",
                                           formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    strainer_parser = subparsers.add_parser('strainer', help="Strain annotations down to genes of interest",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    neighborhood_parser = subparsers.add_parser('neighborhoods', help="Find neighborhoods around genes of interest",
                                                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    merge_annotations_parser = subparsers.add_parser('merge_annotations', help="Merge multiple annotations to one "
                                                                               "larger set",
                                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # parser for annotating mags, you know the real thing
    # TODO: add don't rename flag and give warning that all contig names must be unique
    annotate_parser.add_argument('-i', '--input_fasta', action='append',
                                 help="fasta file, optionally with wildcards to point to multiple fastas",
                                 required=True)
    annotate_parser.add_argument('-o', '--output_dir', help="output directory", required=True)
    annotate_parser.add_argument('--min_contig_size', type=int, default=2500,
                                 help='minimum contig size to be used for gene prediction')
    annotate_parser.add_argument('--config_loc', default=None,
                                       help='location of an alternive config file that will over write the original at run time,'
                                       ' but not be saved or modified')
    prodigal_mode_choices = ['train', 'meta', 'single']
    annotate_parser.add_argument('--prodigal_mode', type=str, default='meta', choices=prodigal_mode_choices,
                                 help='Mode of prodigal to use for gene calling. NOTE: normal or single mode require '
                                      'genomes which are high quality with low contamination and long contigs (average '
                                      'length >3 Kbp).')
    # prodigal_trans_table_choices = ['auto'] + [str(i) for i in range(1, 26)]
    prodigal_trans_table_choices = [str(i) for i in range(1, 26)]
    annotate_parser.add_argument('--trans_table', type=str, default='11', choices=prodigal_trans_table_choices,
                                 help='Translation table for prodigal to use for gene calling.')
    annotate_parser.add_argument('--bit_score_threshold', type=int, default=60,
                                 help='minimum bitScore of search to retain hits')
    annotate_parser.add_argument('--rbh_bit_score_threshold', type=int, default=350,
                                 help='minimum bitScore of reverse best hits to retain hits')
    annotate_parser.add_argument('--kofam_use_dbcan2_thresholds', action='store_true', default=False,
                                 help='Use dbcan2 suggested HMM cutoffs for KOfam annotation instead of KOfam '
                                      'recommended cutoffs. This will be ignored if annotating with KEGG Genes.')
    annotate_parser.add_argument('--custom_db_name', action='append', help="Names of custom databases, can be used"
                                                                           "multiple times.")
    annotate_parser.add_argument('--custom_fasta_loc', action='append', default=[], # empty list is not bug, can't be changed
                                 help="Location of fastas to annotate against, can be used multiple times but"
                                      "must match nubmer of custom_db_name's")
    annotate_parser.add_argument('--custom_hmm_name', action='append',  default=[], # empty list is not bug, can't be changed
                                 help="Names of custom hmm databases, can be used multiple times.")
    annotate_parser.add_argument('--custom_hmm_loc', action='append', default=[], # empty list is not bug, can't be changed
                                 help="Location of hmms to annotate against, can be used multiple times but"
                                      "must match nubmer of custom_hmm_name's")
    annotate_parser.add_argument('--custom_hmm_cutoffs_loc', action='append', default=[], # empty list is not bug, can't be changed
                                 help="Location of file with custom HMM cutoffs and descriptions, can be used "
                                      "multiple times.")
    annotate_parser.add_argument('--gtdb_taxonomy', action='append', default=[], # empty list is not bug, can't be changed
                                 help='Summary file from gtdbtk taxonomy assignment from bins, can be used multiple'
                                      'times')
    annotate_parser.add_argument('--checkm_quality', action='append', default=[], # empty list is not bug, can't be changed
                                 help='Summary of of checkM quality assessment from bins, can be used multiple times')
    annotate_parser.add_argument('--use_uniref', action='store_true', default=False,
                                 help='Annotate these fastas against UniRef, drastically increases run time and memory '
                                      'requirements')
    annotate_parser.add_argument('--use_vogdb', action='store_true', default=False,
                                 help='Annotate these fastas against VOGDB, drastically decreases run time')
    annotate_parser.add_argument('--low_mem_mode', action='store_true', default=False,
                                 help='Skip annotating with uniref and use kofam instead of KEGG genes even if '
                                      'provided. Drastically decreases memory usage')
    annotate_parser.add_argument('--skip_trnascan', action='store_true', default=False)
    annotate_parser.add_argument('--keep_tmp_dir', action='store_true', default=False)
    annotate_parser.add_argument('--threads', type=int, default=10, help='number of processors to use')
    annotate_parser.add_argument('--verbose', action='store_true', default=False)
    annotate_parser.set_defaults(func=annotate_bins)

    # parser for annotating already called genes
    annotate_genes_parser.add_argument('-i', '--input_faa', help="fasta file, optionally with wildcards to point to "
                                                                 "individual MAGs", required=True)
    annotate_genes_parser.add_argument('-o', '--output_dir', help="output directory", required=True)
    annotate_genes_parser.add_argument('--log_file_path', 
                                       help="A name and loctation for the log file")
    annotate_genes_parser.add_argument('--bit_score_threshold', type=int, default=60,
                                       help='minimum bitScore of search to retain hits')
    annotate_genes_parser.add_argument('--rbh_bit_score_threshold', type=int, default=350,
                                       help='minimum bitScore of reverse best hits to retain hits')
    annotate_genes_parser.add_argument('--kofam_use_dbcan2_thresholds', action='store_true', default=False,
                                       help='Use dbcan2 suggested HMM cutoffs for KOfam annotation instead of KOfam '
                                            'recommended cutoffs. This will be ignored if annotating with KEGG Genes.')
    annotate_genes_parser.add_argument('--custom_db_name', action='append', default=[], # empty list is not bug, can't be changed
                                       help="Names of custom databases, can be used multiple times.")
    annotate_genes_parser.add_argument('--custom_fasta_loc', action='append', default=[], # empty list is not bug, can't be changed
                                       help="Location of fastas to annotate against, can be used multiple times but"
                                            "must match nubmer of custom_db_name's")
    annotate_genes_parser.add_argument('--custom_hmm_name', action='append', default=[], # empty list is not bug, can't be changed
                                       help="Names of custom hmm databases, can be used multiple times.")
    annotate_genes_parser.add_argument('--custom_hmm_loc', action='append', default=[], # empty list is not bug, can't be changed
                                       help="Location of hmms to annotate against, can be used multiple times but"
                                            "must match nubmer of custom_hmm_name's")
    annotate_genes_parser.add_argument('--custom_hmm_cutoffs_loc', action='append', default=[], # empty list is not bug, can't be changed
                                       help="Location of file with custom HMM cutoffs and descriptions, can be used "
                                            "multiple times.")
    annotate_genes_parser.add_argument('--use_uniref', action='store_true', default=False,
                                       help='Annotate these fastas against UniRef, drastically increases run time and '
                                            'memory requirements')
    annotate_genes_parser.add_argument('--use_vogdb', action='store_true', default=False,
                                 help='Annotate these fastas against VOGDB, drastically decreases run time')
    annotate_genes_parser.add_argument('--low_mem_mode', action='store_true', default=False,
                                       help='Skip annotating with uniref and use kofam instead of KEGG genes even if '
                                            'provided. Drastically decreases memory usage')
    annotate_genes_parser.add_argument('--config_loc', default=None,
                                       help='location of an alternive config file that will over write the original at run time,'
                                       ' but not be saved or modified')
    annotate_genes_parser.add_argument('--keep_tmp_dir', action='store_true', default=False)
    annotate_genes_parser.add_argument('--threads', type=int, default=10, help='number of processors to use')
    annotate_genes_parser.add_argument('--verbose', action='store_true', default=False)
    annotate_genes_parser.set_defaults(func=annotate_called_genes_cmd)

    # parser for summarizing genomes
    distill_parser.add_argument("-i", "--input_file", help="Annotations path")
    distill_parser.add_argument("-o", "--output_dir", required=True, help="Directory to write summarized genomes")
    distill_parser.add_argument('--log_file_path', 
                                       help="A name and loctation for the log file")
    distill_parser.add_argument("--rrna_path", help="rRNA output from annotation")
    distill_parser.add_argument("--trna_path", help="tRNA output from annotation")
    distill_parser.add_argument("--groupby_column", help="Column from annotations to group as organism units",
                                default='fasta')
    distill_parser.add_argument('--config_loc', default=None,
                                help='location of an alternive config file that will over write the original at run time,'
                                ' but not be saved or modified')
    distill_parser.add_argument("--custom_distillate", help="Custom distillate form to add your own modules")
    distill_parser.add_argument("--distillate_gene_names", action='store_true', default=False,
                                help="Give names of genes instead of counts in genome metabolism summary")
    distill_parser.add_argument("--genomes_per_product", help="Number of genomes per product.html output. Decrease "
                                                              "value if getting JavaScript Error: Maximum call stack "
                                                              "size exceeded when viewing product.html in browser.",
                                default=1000, type=int)
    distill_parser.set_defaults(func=summarize_genomes)

    # parser for getting genes
    input_group = strainer_parser.add_argument_group('Input and output files')
    input_group.add_argument('-i', '--input_tsv', required=True, help="annotations file"
                             " to pull genes from")
    input_group.add_argument('--adjective_sheet', required=False, help='Output tsv file for strainer.'
                             ' Use with the genes faa')
    input_group.add_argument('-f', '--input_fasta', required=True, help='fasta file to filter')
    input_group.add_argument('-o', '--output_fasta', default='pull_genes.fasta',
                             help='location to write filtered fasta')
    name_group = strainer_parser.add_argument_group('Specific names to keep')
    name_group.add_argument('--fastas', nargs='*', default=None, help='space separated list of fastas to keep')
    name_group.add_argument('--scaffolds', nargs='*', default=None,
                            help='space separated list of scaffolds to keep')
    name_group.add_argument('--genes', nargs='*', default=None, help='space separated list of genes to keep')
    annotation_group = strainer_parser.add_argument_group('Annotation filters')
    annotation_group.add_argument('--identifiers', nargs='*', default=None, help='database identifiers to keep')
    annotation_group.add_argument('--categories', nargs='*', default=None,
                                  help='distillate categories to keep genes from')
    annotation_group.add_argument('--custom_distillate', help="Custom distillate form to add your own modules")
    dram_group = strainer_parser.add_argument_group('DRAM based filters')
    dram_group.add_argument('--taxonomy', nargs='*', default=None,
                            help='Level of GTDBTk taxonomy to keep (e.g. c__Clostridia), space separated list')
    dram_group.add_argument('--completeness', default=None, type=float,
                            help='Minimum completeness of genome to keep genes')
    dram_group.add_argument('--contamination', default=None, type=float,
                            help='Maximum contamination of genome to keep genes')
    strainer_parser.set_defaults(func=pull_sequences)

    # parser for getting gene neighborhoods
    neighborhood_parser.add_argument("-i", "--input_file", help="Annotations path")
    neighborhood_parser.add_argument("-o", "--output_dir", required=True, help="Directory to write gene neighborhoods")
    neighborhood_parser.add_argument("--genes", nargs='*', help="Gene names from DRAM to find neighborhoods around")
    neighborhood_parser.add_argument("--identifiers", nargs='*',
                                     help="Database identifiers assigned by DRAM to find neighborhoods around")
    neighborhood_parser.add_argument('--custom_distillate', help="Custom distillate form to add your own modules")
    neighborhood_parser.add_argument("--categories", help="Distillate categories to build gene neighborhoods around.")
    neighborhood_parser.add_argument("--genes_loc", help="Location of genes.fna/genes.faa file to filter to "
                                                         "neighborhoods")
    neighborhood_parser.add_argument("--scaffolds_loc", help="Location of scaffolds.fna file to filter to "
                                                             "neighborhoods")
    neighborhood_parser.add_argument("--distance_genes", type=int, help="Number of genes away from center to include "
                                                                        "in neighborhoods")
    neighborhood_parser.add_argument("--distance_bp", type=int, help="Number of genes away from center to include "
                                                                     "in neighborhoods")
    neighborhood_parser.set_defaults(func=get_gene_neighborhoods)

    # parser for merging annotations
    # TODO: Make it work with append so you can use multiple -i's
    merge_annotations_parser.add_argument("-i", "--input_dirs", help="Path with wildcards pointing to DRAM annotation "
                                                                     "output directories")
    merge_annotations_parser.add_argument("-o", "--output_dir", required=True,  help="Path to output merged annotations files")
    merge_annotations_parser.set_defaults(func=merge_annotations_cmd)

    args = parser.parse_args()
    args_dict = {i: j for i, j in vars(args).items() if i != 'func'}
    args.func(**args_dict)



import pytest
import logging
from mag_annotator.utils import setup_logger

import pandas as pd
import altair as alt

from mag_annotator.summarize_vgfs import (
    filter_to_amgs,
    get_strand_switches,
    make_viral_distillate,
    make_vgf_order,
    make_amg_count_column,
    make_viral_functional_df,
    make_viral_functional_heatmap,
)


@pytest.fixture()
def annotations():
    return pd.DataFrame(
        [
            ["scaffold_1", "K99999", "VTF", 1],
            ["scaffold_1", "K12345", "M", 4],
            ["scaffold_1", "K00001", "MFTJ", 3],
            ["scaffold_1", "K11111", "MKE", 2],
        ],
        index=["gene_1", "gene_2", "gene_3", "gene_4"],
        columns=["scaffold", "ko_id", "amg_flags", "auxiliary_score"],
    )


@pytest.fixture()
def pamgs():
    return pd.DataFrame(
        [["scaffold_1", "K00001", "MFTJ", 3], ["scaffold_1", "K11111", "MKE", 2]],
        index=["gene_3", "gene_4"],
        columns=["scaffold", "ko_id", "amg_flags", "auxiliary_score"],
    )


@pytest.fixture()
def logger(tmpdir):
    logger = logging.getLogger("test_log")
    setup_logger(logger)
    return logger


def test_filter_to_amgs(annotations, pamgs):
    test_pamgs = filter_to_amgs(annotations, remove_transposons=False, max_aux=3)
    pd.testing.assert_frame_equal(test_pamgs, pamgs)


def test_get_strand_switches():
    assert get_strand_switches(["+", "-", "+"]) == 2
    assert get_strand_switches(["+", "+", "+"]) == 0
    assert get_strand_switches(["-"]) == 0
    assert get_strand_switches(["+", "+", "-", "-"]) == 1


@pytest.fixture()
def genome_summary_frame():
    return pd.DataFrame(
        pd.DataFrame(
            [
                ["description", "module1", "main", "header1", "subheader1"],
                ["description2", "module2", "main", "header1", "subheader1"],
                ["description3", "module3", "other", "header1", "subheader1"],
            ],
            index=["K00001", "K12345", "K00001"],
            columns=["gene_description", "module", "sheet", "header", "subheader"],
        )
    )


def test_make_viral_distillate(pamgs, genome_summary_frame, logger):
    amg_database = pd.DataFrame(
        [
            [
                "K11111",
                "gene_4",
                "gene_4",
                "d4",
                "module3",
                "metabolism3",
                "reference2",
                "true",
            ]
        ],
        columns=[
            "KO",
            "EC",
            "PFAM",
            "gene",
            "module",
            "metabolism",
            "reference",
            "verified",
        ],
    )
    output_viral_distillate = make_viral_distillate(
        pamgs, genome_summary_frame, amg_database=amg_database, logger=logger
    )
    expected_viral_distillate = pd.DataFrame(
        [
            [
                "gene_3",
                "scaffold_1",
                "K00001",
                "description",
                "header1",
                "subheader1",
                "module1",
                3,
                "MFTJ",
                "main",
                "genome_summary_form",
                "",
                "",
            ],
            [
                "gene_3",
                "scaffold_1",
                "K00001",
                "description3",
                "header1",
                "subheader1",
                "module3",
                3,
                "MFTJ",
                "other",
                "genome_summary_form",
                "",
                "",
                "",
            ],
            [
                "gene_4",
                "scaffold_1",
                "K11111",
                "d4",
                "",
                "",
                "module3",
                2,
                "MKE",
                "",
                "amg_database",
                "metabolism3",
                "reference2",
                "true",
            ],
        ],
        columns=[
            "gene",
            "scaffold",
            "gene_id",
            "gene_description",
            "header",
            "subheader",
            "module",
            "auxiliary_score",
            "amg_flags",
            "sheet",
            "gene_id_origin",
            "metabolism",
            "reference",
            "verified",
        ],
    )
    output_viral_distillate = output_viral_distillate.sort_index(axis=1).fillna("")
    expected_viral_distillate = expected_viral_distillate.sort_index(axis=1).fillna("")
    pd.testing.assert_frame_equal(output_viral_distillate, expected_viral_distillate)


def test_make_vgf_order(pamgs):
    assert make_vgf_order(pamgs) == ["scaffold_1"]


def test_make_amg_count_column(pamgs):
    assert type(make_amg_count_column(pamgs)) is alt.Chart


@pytest.fixture()
def viral_functional_df():
    return pd.DataFrame(
        [
            ["main", "module1", "gene_3", "K00001", "scaffold_1", True],
            ["other", "module3", "gene_3", "K00001", "scaffold_1", True],
        ],
        columns=[
            "Category",
            "Function",
            "AMG Genes",
            "Genes Present",
            "Contig Name",
            "Present in Contig",
        ],
    )


def test_make_viral_functional_df(pamgs, genome_summary_frame, viral_functional_df):
    test_viral_function_df = make_viral_functional_df(
        pamgs, genome_summary_frame, "scaffold"
    )
    pd.testing.assert_frame_equal(test_viral_function_df, viral_functional_df)


def test_make_viral_functional_heatmap(viral_functional_df):
    assert type(make_viral_functional_heatmap(viral_functional_df)) is alt.HConcatChart
